# -*- coding: utf-8 -*-
import abc

from xclim import indices
from xclim import checks
from xclim.utils import Indicator, Indicator2D

__all__ = ['tn_days_below', 'tx_days_above', 'tx_tn_days_above',
           'heat_wave_frequency', 'heat_wave_max_length', 'heat_wave_index', 'tg_mean', 'tg10p', 'tg90p', 'tn_min',
           'tn_max', 'tn_mean', 'tn10p', 'tn90p', 'tx_min', 'tx_max', 'tx_mean', 'tx10p', 'tx90p',
           'daily_temperature_range', 'daily_temperature_range_variability', 'extreme_temperature_range',
           'cold_spell_duration_index', 'cold_spell_days', 'daily_freezethaw_cycles', 'cooling_degree_days',
           'heating_degree_days', 'growing_degree_days', 'freshet_start', 'frost_days', 'ice_days',
           'consecutive_frost_days', 'growing_season_length', 'tropical_nights']


# TODO: Should we reference the standard vocabulary we're using ?
# E.g. http://vocab.nerc.ac.uk/collection/P07/current/BHMHISG2/


class Tas(Indicator):
    """Class for univariate indices using mean daily temperature as the input."""

    def cfprobe(self, da):
        checks.check_valid(da, 'cell_methods', 'time: mean within days')
        checks.check_valid(da, 'standard_name', 'air_temperature')

    @abc.abstractmethod
    def compute(*args, **kwds):
        """The function computing the indicator."""


class Tasmin(Indicator):
    """Class for univariate indices using min daily temperature as the input."""

    def cfprobe(self, da):
        checks.check_valid(da, 'cell_methods', 'time: minimum within days')
        checks.check_valid(da, 'standard_name', 'air_temperature')

    @abc.abstractmethod
    def compute(*args, **kwds):
        """The function computing the indicator."""


class Tasmax(Indicator):
    """Class for univariate indices using max daily temperature as the input."""

    def cfprobe(self, da):
        checks.check_valid(da, 'cell_methods', 'time: maximum within days')
        checks.check_valid(da, 'standard_name', 'air_temperature')

    @abc.abstractmethod
    def compute(*args, **kwds):
        """The function computing the indicator."""


class TasminTasmax(Indicator2D):

    def cfprobe(self, dan, dax):
        for da in (dan, dax):
            checks.check_valid(da, 'cell_methods', 'time: maximum within days')
            checks.check_valid(da, 'standard_name', 'air_temperature')

    @abc.abstractmethod
    def compute(*args, **kwds):
        """The function computing the indicator."""


tn_days_below = Tasmin(identifier='tn_days_below',
                       units='days',
                       standard_name='number_of_days_with_air_temperature_below_threshold',
                       long_name='Number of days with Tmin < {thresh}',
                       description="{freq} number of days where daily minimum temperature is below {thresh}",
                       cell_methods='time: minimum within days time: sum over days',
                       compute=indices.tn_days_below,
                       )

tx_days_above = Tasmax(identifier='tx_days_above',
                       units='days',
                       standard_name='number_of_days_with_air_temperature_above_threshold',
                       long_name='Number of days with Tmax > {thresh}',
                       description="{freq} number of days where daily maximum temperature exceeds {thresh}",
                       cell_methods='time: maximum within days time: sum over days',
                       compute=indices.tx_days_above,
                       )

tx_tn_days_above = TasminTasmax(identifier='tx_tn_days_above',
                                units='days',
                                standard_name='number_of_days_with_air_temperature_above_threshold',
                                long_name='Number of days with Tmax > {thresh_tasmax} and Tmin > {thresh_tasmin}',
                                description="{freq} number of days where daily maximum temperature exceeds"
                                            " {thresh_tasmax} and minimum temperature exceeds {thresh_tasmin}",
                                cell_methods='',
                                compute=indices.tx_tn_days_above,
                                )

heat_wave_frequency = TasminTasmax(identifier='heat_wave_frequency',
                                   units='',
                                   standard_name='heat_wave_events',
                                   long_name='Number of heat wave events (Tmin > {thresh_tasmin}'
                                             'and Tmax > {thresh_tasmax} for >= {window} days)',
                                   description="{freq} number of heat wave events over a given period. "
                                               "An event occurs when the minimum and maximum daily "
                                               "temperature both exceeds specific thresholds : "
                                               "(Tmin > {thresh_tasmin} and Tmax > {thresh_tasmax}) "
                                               "over a minimum number of days ({window}).",
                                   cell_methods='',
                                   keywords="health,",
                                   compute=indices.heat_wave_frequency,
                                   )

heat_wave_max_length = TasminTasmax(identifier='heat_wave_max_length',
                                    units='days',
                                    standard_name='spell_length_of_days_with_air_temperature_above_threshold',
                                    long_name='Maximum length of heat wave events (Tmin > {thresh_tasmin}'
                                              'and Tmax > {thresh_tasmax} for >= {window} days)',
                                    description="{freq} maximum length of heat wave events occuring in a given period."
                                                "An event occurs when the minimum and maximum daily "
                                                "temperature both exceeds specific thresholds "
                                                "(Tmin > {thresh_tasmin} and Tmax > {thresh_tasmax}) over "
                                                "a minimum number of days ({window}).",
                                    cell_methods='',
                                    keywords="health,",
                                    compute=indices.heat_wave_max_length,
                                    )

heat_wave_index = Tasmax(identifier='heat_wave_index',
                         units='days',
                         standard_name='heat_wave_index',
                         long_name='Number of days that are part of a heatwave',
                         description='{freq} number of days that are part of a heatwave, '
                                     'defined as five or more consecutive days over {thresh}.',
                         cell_methods='',
                         compute=indices.heat_wave_index,
                         )

tg_mean = Tas(identifier='tg_mean',
              units='K',
              standard_name="air_temperature",
              long_name="Mean daily mean temperature",
              description="{freq} mean of daily mean temperature.",
              cell_methods='time: mean within days time: mean over days',
              compute=indices.tg_mean, )

tx_mean = Tasmax(identifier='tx_mean',
                 units='K',
                 standard_name='air_temperature',
                 long_name='Mean daily maximum temperature',
                 description='{freq} mean of daily maximum temperature.',
                 cell_methods='time: maximum within days time: mean over days',
                 compute=indices.tx_mean,
                 )

tx_max = Tasmax(identifier='tx_max',
                units='K',
                standard_name='air_temperature',
                long_name='Maximum daily maximum temperature',
                description='{freq} maximum of daily maximum temperature.',
                cell_methods='time: maximum within days time: maximum over days',
                compute=indices.tx_max,
                )

tx_min = Tasmax(identifier='tx_min',
                units='K',
                standard_name='air_temperature',
                long_name='Minimum daily maximum temperature',
                description='{freq} minimum of daily maximum temperature.',
                cell_methods='time: maximum within days time: minimum over days',
                compute=indices.tx_min,
                )

tn_mean = Tasmin(identifier='tn_mean',
                 units='K',
                 standard_name='air_temperature',
                 long_name='Mean daily minimum temperature',
                 description='{freq} mean of daily minimum temperature.',
                 cell_methods='time: minimum within days time: mean over days',
                 compute=indices.tn_mean,
                 )

tn_max = Tasmin(identifier='tn_max',
                units='K',
                standard_name='air_temperature',
                long_name='Maximum daily minimum temperature',
                description='{freq} maximum of daily minimum temperature.',
                cell_methods='time: minimum within days time: maximum over days',
                compute=indices.tn_max,
                )

tn_min = Tasmin(identifier='tn_min',
                units='K',
                standard_name='air_temperature',
                long_name='Minimum daily minimum temperature',
                description='{freq} minimum of daily minimum temperature.',
                cell_methods='time: minimum within days time: minimum over days',
                compute=indices.tn_min,
                )

daily_temperature_range = TasminTasmax(identifier='dtr',
                                       units='K',
                                       standard_name='air_temperature',
                                       long_name='Mean Diurnal Temperature Range',
                                       description='{freq} mean diurnal temperature range',
                                       cell_methods='time range within days time: mean over days',
                                       compute=indices.daily_temperature_range,
                                       )

daily_temperature_range_variability = TasminTasmax(identifier='dtrvar',
                                                   units='K',
                                                   standard_name='air_temperature',
                                                   long_name='Mean Diurnal Temperature Range Variability',
                                                   description='{freq} mean diurnal temparature range variability ('
                                                               'defined as the average day-to-day variation '
                                                               'in daily temperature range '
                                                               'for the given time period)',
                                                   cell_methods='time range within days time: difference '
                                                                'over days time: mean over days',
                                                   compute=indices.daily_temperature_range_variability,
                                                   )

extreme_temperature_range = TasminTasmax(identifier='etr',
                                         units='K',
                                         standard_name='air_temperature',
                                         long_name='Intra-period Extreme Temperature Range',

                                         description='{freq} range between the maximum of daily max temperature '
                                                     '(tx_max) and the minimum of daily min temperature (tn_min)',
                                         compute=indices.extreme_temperature_range,
                                         )

cold_spell_duration_index = Tasmin(identifier='cold_spell_duration_index',
                                   var_name='csdi_{window}',
                                   units='days',
                                   standard_name='cold_spell_duration_index',
                                   long_name='Cold Spell Duration Index, count of days with at '
                                             'least {window} consecutive days when Tmin < 10th percentile',
                                   description='{freq} number of days with at least {window} consecutive days'
                                              ' where the daily minimum temperature is below the 10th '
                                              'percentile. The 10th percentile should be computed for '
                                              'a 5-day window centred on each calendar day in the  1961-1990 period',
                                   cell_methods='',
                                   compute=indices.cold_spell_duration_index,
                                   )

cold_spell_days = Tas(identifier='cold_spell_days',
                      units='days',
                      standard_name='cold_spell_days',
                      long_name='cold spell index',
                      description='{freq} number of days that are part of a cold spell, defined as {window} '
                                  'or more consecutive days with mean daily '
                                  'temperature below  {thresh}.',
                      cell_methods='',
                      compute=indices.cold_spell_days,
                      )

daily_freezethaw_cycles = TasminTasmax(identifier='dlyfrzthw',
                                       units='days',
                                       standard_name='daily_freezethaw_cycles',
                                       long_name='daily freezethaw cycles',
                                       description='{freq} number of days with a diurnal freeze-thaw cycle '
                                                   ': Tmax > 0℃ and Tmin < 0℃.',
                                       cell_methods='',
                                       compute=indices.daily_freezethaw_cycles,
                                       )

cooling_degree_days = Tas(identifier='cooling_degree_days',
                          units='K days',
                          standard_name='integral_of_air_temperature_excess_wrt_time',
                          long_name='Cooling Degree Days (Tmean > {thresh})',
                          description='{freq} cooling degree days above {thresh}.',
                          cell_methods='time: mean within days time: sum over days',
                          compute=indices.cooling_degree_days,
                          )

heating_degree_days = Tas(identifier='heating_degree_days',
                          units='K days',
                          standard_name='integral_of_air_temperature_deficit_wrt_time',
                          long_name='Heating Degree Days (Tmean < {thresh})',
                          description='{freq} heating degree days below {thresh}.',
                          cell_methods='time: mean within days time: sum over days',
                          compute=indices.heating_degree_days,
                          )

growing_degree_days = Tas(identifier='growing_degree_days',
                          units='K days',
                          standard_name='integral_of_air_temperature_excess_wrt_time',
                          long_name='growing degree days above {thresh}',
                          description='{freq} growing degree days above {thresh}',
                          cell_methods='time: mean within days time: sum over days',
                          compute=indices.growing_degree_days,
                          )

freshet_start = Tas(identifier='freshet_start',
                    units='',
                    standard_name='day_of_year',
                    long_name="Day of year of spring freshet start",
                    description="Day of year of spring freshet start, defined as the first day a temperature "
                                "threshold of {thresh} is exceeded for at least {window} days.",
                    compute=indices.freshet_start)

frost_days = Tasmin(identifier='frost_days',
                    units='days',
                    standard_name='days_with_air_temperature_below_threshold',
                    long_name='Number of Frost Days (Tmin < 0C)',
                    description='{freq} number of days with minimum daily '
                                'temperature below 0℃.',
                    cell_methods='time: minimum within days time: sum over days',
                    compute=indices.frost_days,
                    )

ice_days = Tasmax(identifier='ice_days',
                  standard_name='days_with_air_temperature_below_threshold',
                  units='days',
                  long_name='Number of Ice Days (Tmax < 0℃)',
                  description='{freq} number of days with maximum daily '
                              'temperature below 0℃',
                  cell_methods='time: maximum within days time: sum over days',
                  compute=indices.ice_days,
                  )

consecutive_frost_days = Tasmin(identifier='consecutive_frost_days',
                                units='days',
                                standard_name='spell_length_of_days_with_air_temperature_below_threshold',
                                long_name='Maximum number of consecutive days with Tmin < 0C',
                                description='{freq} maximum number of consecutive days with '
                                            'minimum daily temperature below 0℃',
                                cell_methods='time: min within days time: maximum over days',
                                compute=indices.consecutive_frost_days,
                                )

growing_season_length = Tas(identifier='growing_season_length',
                            units='days',
                            standard_name='growing_season_length',
                            long_name='ETCCDI Growing Season Length (Tmean > {thresh})',
                            description='{freq} number of days between the first occurrence of at least '
                                        'six consecutive days with mean daily temperature over {thresh} and '
                                        'the first occurrence of at least {window} consecutive days with '
                                        'mean daily temperature below {thresh} after July 1st in the northern '
                                        'hemisphere and January 1st in the southern hemisphere',
                            cell_methods='',
                            compute=indices.growing_season_length,
                            )

tropical_nights = Tasmin(identifier='tropical_nights',
                         units='days',
                         standard_name='number_of_days_with_air_temperature_above_threshold',
                         long_name='Number of Tropical Nights (Tmin > {thresh})',
                         description='{freq} number of Tropical Nights : defined as days with minimum daily temperature'
                                     ' above {thresh}',
                         cell_methods='time: minimum within days time: sum over days',
                         compute=indices.tropical_nights,
                         )

tg90p = Tas(identifier='tg90p',
            units='days',
            standard_name='days_with_air_temperature_above_threshold',
            long_name='Number of days when Tmean > 90th percentile',
            description='{freq} number of days with mean daily temperature above the 90th percentile.'
                        'The 90th percentile is to be computed for a 5 day window centered on each calendar day '
                        'for a reference period.',
            cell_methods='time: mean within days time: sum over days',
            compute=indices.tg90p,
            )

tg10p = Tas(identifier='tg10p',
            units='days',
            standard_name='days_with_air_temperature_below_threshold',
            long_name='Number of days when Tmean < 10th percentile',
            description='{freq} number of days with mean daily temperature below the 10th percentile.'
                        'The 10th percentile is to be computed for a 5 day window centered on each calendar day '
                        'for a reference period.',
            cell_methods='time: mean within days time: sum over days',
            compute=indices.tg10p
            )

tx90p = Tasmax(identifier='tx90p',
               units='days',
               standard_name='days_with_air_temperature_above_threshold',
               long_name='Number of days when Tmax > 90th percentile',
               description='{freq} number of days with maximum daily temperature above the 90th percentile.'
                           'The 90th percentile is to be computed for a 5 day window centered on each calendar day '
                           'for a reference period.',
               cell_methods='time: maximum within days time: sum over days',
               compute=indices.tx90p,
               )

tx10p = Tasmax(identifier='tx10p',
               units='days',
               standard_name='days_with_air_temperature_below_threshold',
               long_name='Number of days when Tmax < 10th percentile',
               description='{freq} number of days with maximum daily temperature below the 10th percentile.'
                           'The 10th percentile is to be computed for a 5 day window centered on each calendar day '
                           'for a reference period.',
               cell_methods='time: maximum within days time: sum over days',
               compute=indices.tx10p
               )

tn90p = Tasmin(identifier='tn90p',
               units='days',
               standard_name='days_with_air_temperature_above_threshold',
               long_name='Number of days when Tmin > 90th percentile',
               description='{freq} number of days with minimum daily temperature above the 90th percentile.'
                           'The 90th percentile is to be computed for a 5 day window centered on each calendar day '
                           'for a reference period.',
               cell_methods='time: minimum within days time: sum over days',
               compute=indices.tn90p,
               )

tn10p = Tasmin(identifier='tn10p',
               units='days',
               standard_name='days_with_air_temperature_below_threshold',
               long_name='Number of days when Tmin < 10th percentile',
               description='{freq} number of days with minimum daily temperature below the 10th percentile.'
                           'The 10th percentile is to be computed for a 5 day window centered on each calendar day '
                           'for a reference period.',
               cell_methods='time: minimum within days time: sum over days',
               compute=indices.tn10p
               )



# -*- coding: utf-8 -*-

"""Top-level package for xclim."""

from functools import partial

from xclim import indices
import sys

# from .stats import fit, test

__author__ = """Travis Logan"""
__email__ = 'logan.travis@ouranos.ca'
__version__ = '0.10.2-beta'


def build_module(name, objs, doc='', source=None, mode='ignore'):
    """Create a module from imported objects.

    Parameters
    ----------
    name : str
      New module name.
    objs : dict
      Dictionary of the objects (or their name) to import into the module,
      keyed by the name they will take in the created module.
    doc : str
      Docstring of the new module.
    source : Module object
      Module where objects are defined if not explicitly given.
    mode : {'raise', 'warn', 'ignore'}
      How to deal with missing objects.


    Returns
    -------
    ModuleType
      A module built from a list of objects' name.

    """
    import types
    import warnings
    import logging

    logging.captureWarnings(capture=True)

    try:
        out = types.ModuleType(name, doc)
    except TypeError:
        msg = "Module '{}' is not properly formatted".format(name)
        raise TypeError(msg)

    for key, obj in objs.items():
        if isinstance(obj, str) and source is not None:
            module_mappings = getattr(source, obj, None)
        else:
            module_mappings = obj

        if module_mappings is None:
            msg = "{} has not been implemented.".format(obj)
            if mode == 'raise':
                raise NotImplementedError(msg)
            elif mode == 'warn':
                warnings.warn(msg)
            else:
                logging.info(msg)

        else:
            out.__dict__[key] = module_mappings
            try:
                module_mappings.__module__ = name
            except AttributeError:
                msg = "{} is not a function".format(module_mappings)
                raise AttributeError(msg)

    sys.modules[name] = out
    return out


def __build_icclim(mode='warn'):

    #  ['TG', 'TX', 'TN', 'TXx', 'TXn', 'TNx', 'TNn', 'SU', 'TR', 'CSU', 'GD4', 'FD', 'CFD', 'GSL',
    #   'ID', 'HD17', 'CDD', 'CWD', 'PRCPTOT', 'RR1', 'SDII', 'R10mm', 'R20mm', 'RX1day', 'RX5day',
    #   'SD', 'SD1', 'SD5cm', 'SD50cm', 'DTR', 'ETR', 'vDTR', 'TG10p', 'TX10p', 'TN10p', 'TG90p',
    #   'TX90p', 'TN90p', 'WSDI', 'CSDI', 'R75p', 'R75pTOT', 'R95p', 'R95pTOT', 'R99p', 'R99pTOT']

    # Use partials to specify default value ?
    # TODO : Complete mappings for ICCLIM indices
    mapping = {'TG': indices.tg_mean,
               'TX': indices.tx_mean,
               'TN': indices.tn_mean,
               'TG90p': indices.tg90p,
               'TG10p': indices.tg10p,
               'TGx': indices.tg_max,
               'TGn': indices.tg_min,
               'TX90p': indices.tx90p,
               'TX10p': indices.tx10p,
               'TXx': indices.tx_max,
               'TXn': indices.tx_min,
               'TN90p': indices.tn90p,
               'TN10p': indices.tn10p,
               'TNx': indices.tn_max,
               'TNn': indices.tn_min,
               'SU': indices.tx_days_above,
               'TR': indices.tropical_nights,
               # 'CSU': None,
               'GD4': partial(indices.growing_degree_days, thresh=4),
               'FD': indices.frost_days,
               'CFD': indices.consecutive_frost_days,
               'GSL': indices.growing_season_length,
               'ID': indices.ice_days,
               'HD17': indices.heating_degree_days,
               'CDD': indices.maximum_consecutive_dry_days,
               'CWD': indices.maximum_consecutive_wet_days,
               'PRCPTOT': indices.precip_accumulation,
               'RR1': indices.wetdays,
               # 'SDII': None,
               'ETR': indices.extreme_temperature_range,
               'DTR': indices.daily_temperature_range,
               'vDTR': indices.daily_temperature_range_variability,
               # 'R10mm': None,
               # 'R20mm': None,
               # 'RX1day': None,
               # 'RX5day': None,
               # 'R75p': None,
               # 'R75pTOT':None,
               # 'R95p': None,
               # 'R95pTOT': None,
               # 'R99p': None,
               # 'R99pTOT': None,
               # 'SD': None,
               # 'SD1': None,
               # 'SD5cm': None,
               # 'SD50cm': None,
               }

    mod = build_module('xclim.icclim', mapping, doc="""ICCLIM indices""", mode=mode)
    return mod


icclim = __build_icclim('ignore')



# -*- coding: utf-8 -*-
import numpy as np

from xclim.indices import base_flow_index
from xclim import checks
from xclim.utils import Indicator, wrapped_partial
from xclim import generic
# from boltons.funcutils import FunctionBuilder
# import calendar


class Streamflow(Indicator):
    units = 'm^3 s-1'
    context = 'hydro'
    standard_name = 'discharge'

    @staticmethod
    def compute(*args, **kwds):
        pass

    def cfprobe(self, q):
        checks.check_valid(q, 'standard_name', 'streamflow')


class Stats(Streamflow):
    def missing(self, *args, **kwds):
        """Return whether an output is considered missing or not."""
        from functools import reduce

        indexer = kwds['indexer']
        freq = kwds['freq'] or generic.default_freq(**indexer)

        miss = (checks.missing_any(da, freq, **indexer) for da in args)
        return reduce(np.logical_or, miss)


# Disable the missing value check because the output here is not a time series.
class FA(Streamflow):
    def missing(self, *args, **kwds):
        """Return whether an output is considered missing or not."""
        return False


# Disable the daily checks because the inputs are period extremas.
class Fit(FA):
    def validate(self, da):
        pass


base_flow_index = Streamflow(identifier='base_flow_index',
                             units='',
                             long_name="Base flow index",
                             compute=base_flow_index)


freq_analysis = FA(identifier='freq_analysis',
                   var_name='q{window}{mode}{indexer}',
                   long_name='N-year return period {mode} {indexer} {window}-day flow',
                   description="Streamflow frequency analysis for the {mode} {indexer} {window}-day flow "
                               "estimated using the {dist} distribution.",
                   compute=generic.frequency_analysis)


stats = Stats(identifier='stats',
              var_name='q{indexer}{op}',
              long_name='{freq} {op} of {indexer} daily flow ',
              description="{freq} {op} of {indexer} daily flow",
              compute=generic.select_resample_op)


fit = Fit(identifier='fit',
          var_name='params',
          units='',
          standard_name='{dist} parameters',
          long_name="{dist} distribution parameters",
          description="Parameters of the {dist} distribution fitted ",
          cell_methods='time: fit',
          compute=generic.fit)


doy_qmax = Streamflow(identifier='doy_qmax',
                      var_name='q{indexer}_doy_qmax',
                      long_name='Day of the year of the maximum over {indexer}',
                      description='Day of the year of the maximum over {indexer}',
                      units='',
                      _partial=True,
                      compute=wrapped_partial(generic.select_resample_op, op=generic.doymax))


doy_qmin = Streamflow(identifier='doy_qmin',
                      var_name='q{indexer}_doy_qmin',
                      long_name='Day of the year of the minimum over {indexer}',
                      description='Day of the year of the minimum over {indexer}',
                      units='',
                      _partial=True,
                      compute=wrapped_partial(generic.select_resample_op, op=generic.doymin))



import os
import sys

import numpy as np
import pandas as pd
import pytest
import xarray as xr
import xclim.atmos as atmos

TESTS_HOME = os.path.abspath(os.path.dirname(__file__))
TESTS_DATA = os.path.join(TESTS_HOME, 'testdata')
K2C = 273.15


class TestRainOnFrozenGround():
    nc_pr = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_pr_1990.nc')
    nc_tasmax = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_tasmax_1990.nc')
    nc_tasmin = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_tasmin_1990.nc')

    def test_3d_data_with_nans(self):
        pr = xr.open_dataset(self.nc_pr).pr
        prMM = pr.copy()
        prMM.values *= 86400.
        prMM.attrs['units'] = 'mm/day'

        tasmax = xr.open_dataset(self.nc_tasmax).tasmax
        tasmin = xr.open_dataset(self.nc_tasmin).tasmin
        tas = 0.5 * (tasmax + tasmin)
        tas.attrs = tasmax.attrs
        tasC = tas.copy()
        tasC.values -= K2C
        tasC.attrs['units'] = 'C'

        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        out1 = atmos.rain_on_frozen_ground_days(pr, tas, freq='MS')
        out2 = atmos.rain_on_frozen_ground_days(prMM, tas, freq='MS')
        out3 = atmos.rain_on_frozen_ground_days(prMM, tasC, freq='MS')
        out4 = atmos.rain_on_frozen_ground_days(pr, tasC, freq='MS')
        pr.attrs['units'] = 'kg m-2 s-1'
        out5 = atmos.rain_on_frozen_ground_days(pr, tas, freq='MS')
        out6 = atmos.rain_on_frozen_ground_days(pr, tasC, freq='MS')
        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)
        np.testing.assert_array_equal(out1, out4)
        np.testing.assert_array_equal(out1, out5)
        np.testing.assert_array_equal(out1, out6)

        assert np.isnan(out1.values[0, 1, 0])

        assert np.isnan(out1.values[0, -1, -1])

        # synthetic data
        tas1 = tas[0:31, 47, 8] * 0 + K2C - 1
        tas1.attrs = tas.attrs
        pr1 = pr[0:31, 47, 8] * 0 + 25
        pr1.attrs = pr.attrs
        tas1[10] += 5
        tas1[20] += 5

        rfrz = atmos.rain_on_frozen_ground_days(pr1, tas1, freq='MS')

        np.testing.assert_array_equal(rfrz, 2)


class TestPrecipAccumulation():
    # TODO: replace by fixture
    nc_file = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_pr_1990.nc')

    def test_3d_data_with_nans(self):
        # test with 3d data
        pr = xr.open_dataset(self.nc_file).pr  # mm/s
        prMM = xr.open_dataset(self.nc_file).pr
        prMM *= 86400
        prMM.attrs['units'] = 'mm/day'
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        out1 = atmos.precip_accumulation(pr, freq='MS')
        out2 = atmos.precip_accumulation(prMM, freq='MS')

        # test kg m-2 s-1
        pr.attrs['units'] = 'kg m-2 s-1'
        out3 = atmos.precip_accumulation(pr, freq='MS')

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0].values

        prTot = x1.sum()

        np.testing.assert_almost_equal(prTot, out1.values[0, 0, 0], 4)

        assert np.isnan(out1.values[0, 1, 0])

        assert np.isnan(out1.values[0, -1, -1])


class TestWetDays():
    # TODO: replace by fixture
    nc_file = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_pr_1990.nc')

    def test_3d_data_with_nans(self):
        # test with 3d data
        pr = xr.open_dataset(self.nc_file).pr
        prMM = xr.open_dataset(self.nc_file).pr
        prMM.values *= 86400.
        prMM.attrs['units'] = 'mm/day'
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        pr_min = '5 mm/d'
        out1 = atmos.wetdays(pr, thresh=pr_min, freq='MS')
        out2 = atmos.wetdays(prMM, thresh=pr_min, freq='MS')

        # test kg m-2 s-1
        pr.attrs['units'] = 'kg m-2 s-1'
        out3 = atmos.wetdays(pr, thresh=pr_min, freq='MS')

        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0].values

        wd1 = (x1 >= int(pr_min.split(' ')[0])).sum()

        assert wd1 == out1.values[0, 0, 0]

        assert np.isnan(out1.values[0, 1, 0])

        # make sure that vector with all nans gives nans whatever skipna
        assert np.isnan(out1.values[0, -1, -1])
        # assert (np.isnan(wds.values[0, -1, -1]))


class TestDailyIntensity():
    # testing of wet_day and daily_pr_intensity, both are related

    nc_file = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_pr_1990.nc')

    def test_3d_data_with_nans(self):
        # test with 3d data
        pr = xr.open_dataset(self.nc_file).pr
        prMM = xr.open_dataset(self.nc_file).pr
        prMM.values *= 86400.
        prMM.attrs['units'] = 'mm/day'
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        # compute with both skipna options
        pr_min = '2 mm/d'
        # dis = daily_pr_intensity(pr, pr_min=pr_min, freq='MS', skipna=True)

        out1 = atmos.daily_pr_intensity(pr, thresh=pr_min, freq='MS')
        out2 = atmos.daily_pr_intensity(prMM, thresh=pr_min, freq='MS')

        # test kg m-2 s-1
        pr.attrs['units'] = 'kg m-2 s-1'
        out3 = atmos.daily_pr_intensity(pr, thresh=pr_min, freq='MS')

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3, 3)

        x1 = prMM[:31, 0, 0].values

        di1 = x1[x1 >= int(pr_min.split(' ')[0])].mean()
        # buffer = np.ma.masked_invalid(x2)
        # di2 = buffer[buffer >= pr_min].mean()

        assert np.allclose(di1, out1.values[0, 0, 0])
        # assert (np.allclose(di1, dis.values[0, 0, 0]))
        assert np.isnan(out1.values[0, 1, 0])
        # assert (np.allclose(di2, dis.values[0, 1, 0]))
        assert np.isnan(out1.values[0, -1, -1])
        # assert (np.isnan(dis.values[0, -1, -1]))


class TestMax1Day():
    # testing of wet_day and daily_pr_intensity, both are related

    nc_file = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_pr_1990.nc')

    def test_3d_data_with_nans(self):
        # test with 3d data
        pr = xr.open_dataset(self.nc_file).pr
        prMM = xr.open_dataset(self.nc_file).pr
        prMM.values *= 86400.
        prMM.attrs['units'] = 'mm/day'
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        out1 = atmos.max_1day_precipitation_amount(pr, freq='MS')
        out2 = atmos.max_1day_precipitation_amount(prMM, freq='MS')

        # test kg m-2 s-1
        pr.attrs['units'] = 'kg m-2 s-1'
        out3 = atmos.max_1day_precipitation_amount(pr, freq='MS')

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3, 3)

        x1 = prMM[:31, 0, 0].values
        rx1 = x1.max()

        assert np.allclose(rx1, out1.values[0, 0, 0])
        # assert (np.allclose(di1, dis.values[0, 0, 0]))
        assert np.isnan(out1.values[0, 1, 0])
        # assert (np.allclose(di2, dis.values[0, 1, 0]))
        assert np.isnan(out1.values[0, -1, -1])
        # assert (np.isnan(dis.values[0, -1, -1]))


class TestMaxNDay():
    # testing of wet_day and daily_pr_intensity, both are related

    nc_file = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_pr_1990.nc')

    def test_3d_data_with_nans(self):
        # test with 3d data
        pr = xr.open_dataset(self.nc_file).pr
        prMM = xr.open_dataset(self.nc_file).pr
        prMM.values *= 86400.
        prMM.attrs['units'] = 'mm/day'
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        wind = 3
        out1 = atmos.max_n_day_precipitation_amount(pr, window=wind, freq='MS')
        out2 = atmos.max_n_day_precipitation_amount(prMM, window=wind, freq='MS')

        # test kg m-2 s-1
        pr.attrs['units'] = 'kg m-2 s-1'
        out3 = atmos.max_n_day_precipitation_amount(pr, window=wind, freq='MS')

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3, 3)

        x1 = prMM[:31, 0, 0].values
        df = pd.DataFrame({'pr': x1})
        rx3 = df.rolling(wind).sum().max()

        assert np.allclose(rx3, out1.values[0, 0, 0])

        assert np.isnan(out1.values[0, 1, 0])

        assert np.isnan(out1.values[0, -1, -1])


@pytest.mark.skipif(sys.version_info < (3, 5), reason="too slow to evaluate on python2.7")
class TestMaxConsecWetDays():
    # TODO: replace by fixture
    nc_file = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_pr_1990.nc')

    def test_3d_data_with_nans(self):
        # test with 3d data
        pr = xr.open_dataset(self.nc_file).pr
        prMM = xr.open_dataset(self.nc_file).pr
        prMM.values *= 86400.
        prMM.attrs['units'] = 'mm/day'
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        pr_min = '5 mm/d'
        out1 = atmos.maximum_consecutive_wet_days(pr, thresh=pr_min, freq='MS')
        out2 = atmos.maximum_consecutive_wet_days(prMM, thresh=pr_min, freq='MS')

        # test kg m-2 s-1
        pr.attrs['units'] = 'kg m-2 s-1'
        out3 = atmos.maximum_consecutive_wet_days(pr, thresh=pr_min, freq='MS')

        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0] * 0.
        x1[5:10] = 10
        x1.attrs['units'] = 'mm/day'
        cwd1 = atmos.maximum_consecutive_wet_days(x1, freq='MS')

        assert cwd1 == 5

        assert np.isnan(out1.values[0, 1, 0])

        # make sure that vector with all nans gives nans whatever skipna
        assert np.isnan(out1.values[0, -1, -1])
        # assert (np.isnan(wds.values[0, -1, -1]))


@pytest.mark.skipif(sys.version_info < (3, 5), reason="too slow to evaluate on python2.7")
class TestMaxConsecDryDays():
    # TODO: replace by fixture
    nc_file = os.path.join(TESTS_DATA, 'NRCANdaily', 'nrcan_canada_daily_pr_1990.nc')

    def test_3d_data_with_nans(self):
        # test with 3d data
        pr = xr.open_dataset(self.nc_file).pr
        prMM = xr.open_dataset(self.nc_file).pr
        prMM.values *= 86400.
        prMM.attrs['units'] = 'mm/day'
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        pr_min = '5 mm/d'
        out1 = atmos.maximum_consecutive_dry_days(pr, thresh=pr_min, freq='MS')
        out2 = atmos.maximum_consecutive_dry_days(prMM, thresh=pr_min, freq='MS')

        # test kg m-2 s-1
        pr.attrs['units'] = 'kg m-2 s-1'
        out3 = atmos.maximum_consecutive_dry_days(pr, thresh=pr_min, freq='MS')

        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0] * 0. + 50.0
        x1[5:10] = 0
        x1.attrs['units'] = 'mm/day'
        cdd1 = atmos.maximum_consecutive_dry_days(x1, freq='MS')

        assert cdd1 == 5

        assert np.isnan(out1.values[0, 1, 0])

        # make sure that vector with all nans gives nans whatever skipna
        assert np.isnan(out1.values[0, -1, -1])
        # assert (np.isnan(wds.values[0, -1, -1]))



import logging

import numpy as np
import xarray as xr

from xclim import utils, run_length as rl
from xclim.utils import declare_units, units

# logging.basicConfig(level=logging.DEBUG)
# logging.captureWarnings(True)

xr.set_options(enable_cftimeindex=True)  # Set xarray to use cftimeindex

# Frequencies : YS: year start, QS-DEC: seasons starting in december, MS: month start
# See http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases

# -------------------------------------------------- #
# ATTENTION: ASSUME ALL INDICES WRONG UNTIL TESTED ! #
# -------------------------------------------------- #

__all__ = ['cold_spell_days', 'daily_pr_intensity', 'maximum_consecutive_wet_days', 'cooling_degree_days',
           'freshet_start', 'growing_degree_days', 'growing_season_length', 'heat_wave_index', 'heating_degree_days',
           'tn_days_below', 'tx_days_above', 'warm_day_frequency', 'warm_night_frequency', 'wetdays',
           'maximum_consecutive_dry_days', 'max_n_day_precipitation_amount', 'tropical_nights']


@declare_units('days', tas='[temperature]', thresh='[temperature]')
def cold_spell_days(tas, thresh='-10 degC', window=5, freq='AS-JUL'):
    r"""Cold spell days

    The number of days that are part of a cold spell, defined as five or more consecutive days with mean daily
    temperature below a threshold in °C.

    Parameters
    ----------
    tas : xarrray.DataArray
      Mean daily temperature [℃] or [K]
    thresh : str
      Threshold temperature below which a cold spell begins [℃] or [K]. Default : '-10 degC'
    window : int
      Minimum number of days with temperature below threshold to qualify as a cold spell.
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Cold spell days.

    Notes
    -----
    Let :math:`T_i` be the mean daily temperature on day :math:`i`, the number of cold spell days during
    period :math:`\phi` is given by

    .. math::

       \sum_{i \in \phi} \prod_{j=i}^{i+5} [T_j < thresh]

    where :math:`[P]` is 1 if :math:`P` is true, and 0 if false.

    """
    t = utils.convert_units_to(thresh, tas)
    over = tas < t
    group = over.resample(time=freq)

    return group.apply(rl.windowed_run_count, window=window, dim='time')


@declare_units('mm/day', pr='[precipitation]', thresh='[precipitation]')
def daily_pr_intensity(pr, thresh='1 mm/day', freq='YS'):
    r"""Average daily precipitation intensity

    Return the average precipitation over wet days.

    Parameters
    ----------
    pr : xarray.DataArray
      Daily precipitation [mm/d or kg/m²/s]
    thresh : str
      precipitation value over which a day is considered wet. Default : '1 mm/day'
    freq : str, optional
      Resampling frequency defining the periods
      defined in http://pandas.pydata.org/pandas-docs/stable/timeseries.html#resampling. Default : '1 mm/day'

    Returns
    -------
    xarray.DataArray
      The average precipitation over wet days for each period

    Notes
    -----
    Let :math:`\mathbf{p} = p_0, p_1, \ldots, p_n` be the daily precipitation and :math:`thresh` be the precipitation
    threshold defining wet days. Then the daily precipitation intensity is defined as

    .. math::

       \frac{\sum_{i=0}^n p_i [p_i \leq thresh]}{\sum_{i=0}^n [p_i \leq thresh]}

    where :math:`[P]` is 1 if :math:`P` is true, and 0 if false.

    Examples
    --------
    The following would compute for each grid cell of file `pr.day.nc` the average
    precipitation fallen over days with precipitation >= 5 mm at seasonal
    frequency, ie DJF, MAM, JJA, SON, DJF, etc.:

    >>> pr = xr.open_dataset('pr.day.nc')
    >>> daily_int = daily_pr_intensity(pr, thresh='5 mm/day', freq="QS-DEC")

    """
    t = utils.convert_units_to(thresh, pr, 'hydro')

    # put pr=0 for non wet-days
    pr_wd = xr.where(pr >= t, pr, 0)
    pr_wd.attrs['units'] = pr.units

    # sum over wanted period
    s = pr_wd.resample(time=freq).sum(dim='time', keep_attrs=True)
    sd = utils.pint_multiply(s, 1 * units.day, 'mm')

    # get number of wetdays over period
    wd = wetdays(pr, thresh=thresh, freq=freq)
    return sd / wd


@declare_units('days', pr='[precipitation]', thresh='[precipitation]')
def maximum_consecutive_wet_days(pr, thresh='1 mm/day', freq='YS'):
    r"""Consecutive wet days.

    Returns the maximum number of consecutive wet days.

    Parameters
    ---------
    pr : xarray.DataArray
      Mean daily precipitation flux [Kg m-2 s-1] or [mm]
    thresh : str
      Threshold precipitation on which to base evaluation [Kg m-2 s-1] or [mm]. Default : '1 mm/day'
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      The maximum number of consecutive wet days.

    Notes
    -----
    Let :math:`\mathbf{x}=x_0, x_1, \ldots, x_n` be a daily precipitation series and
    :math:`\mathbf{s}` be the sorted vector of indices :math:`i` where :math:`[p_i > thresh] \neq [p_{i+1} >
    thresh]`, that is, the days when the precipitation crosses the *wet day* threshold.
    Then the maximum number of consecutive wet days is given by

    .. math::


       \max(\mathbf{d}) \quad \mathrm{where} \quad d_j = (s_j - s_{j-1}) [x_{s_j} > 0\celsius]

    where :math:`[P]` is 1 if :math:`P` is true, and 0 if false. Note that this formula does not handle sequences at
    the start and end of the series, but the numerical algorithm does.
    """
    thresh = utils.convert_units_to(thresh, pr, 'hydro')

    group = (pr > thresh).resample(time=freq)
    return group.apply(rl.longest_run, dim='time')


@declare_units('C days', tas='[temperature]', thresh='[temperature]')
def cooling_degree_days(tas, thresh='18 degC', freq='YS'):
    r"""Cooling degree days

    Sum of degree days above the temperature threshold at which spaces are cooled.

    Parameters
    ----------
    tas : xarray.DataArray
      Mean daily temperature [℃] or [K]
    thresh : str
      Temperature threshold above which air is cooled. Default : '18 degC'
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Cooling degree days

    Notes
    -----
    Let :math:`x_i` be the daily mean temperature at day :math:`i`. Then the cooling degree days above
    temperature threshold :math:`thresh` over period :math:`\phi` is given by:

    .. math::

        \sum_{i \in \phi} (x_{i}-{thresh} [x_i > thresh]

    where :math:`[P]` is 1 if :math:`P` is true, and 0 if false.
    """
    thresh = utils.convert_units_to(thresh, tas)

    return tas.pipe(lambda x: x - thresh) \
        .clip(min=0) \
        .resample(time=freq) \
        .sum(dim='time')


@declare_units('', tas='[temperature]', thresh='[temperature]')
def freshet_start(tas, thresh='0 degC', window=5, freq='YS'):
    r"""First day consistently exceeding threshold temperature.

    Returns first day of period where a temperature threshold is exceeded
    over a given number of days.

    Parameters
    ----------
    tas : xarray.DataArray
      Mean daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to base evaluation [℃] or [K]. Default '0 degC'
    window : int
      Minimum number of days with temperature above threshold needed for evaluation
    freq : str, optional
      Resampling frequency

    Returns
    -------
    float
      Day of the year when temperature exceeds threshold over a given number of days for the first time. If there are
      no such day, return np.nan.

    Notes
    -----
    Let :math:`x_i` be the daily mean temperature at day of the year :math:`i` for values of :math:`i` going from 1
    to 365 or 366. The start date of the freshet is given by the smallest index :math:`i` for which

    .. math::

       \prod_{j=i}^{i+w} [x_j > thresh]

    is true, where :math:`w` is the number of days the temperature threshold should be exceeded,  and :math:`[P]` is
    1 if :math:`P` is true, and 0 if false.
    """
    thresh = utils.convert_units_to(thresh, tas)
    over = (tas > thresh)
    group = over.resample(time=freq)
    return group.apply(rl.first_run_ufunc, window=window, index='dayofyear')


@declare_units('C days', tas='[temperature]', thresh='[temperature]')
def growing_degree_days(tas, thresh='4.0 degC', freq='YS'):
    r"""Growing degree-days over threshold temperature value [℃].

    The sum of degree-days over the threshold temperature.

    Parameters
    ---------
    tas : xarray.DataArray
      Mean daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to base evaluation [℃] or [K]. Default: '4.0 degC'.
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      The sum of growing degree-days above 4℃

    Notes
    -----
    Let :math:`TG_{ij}` be the daily mean temperature at day :math:`i` of period :math:`j`. Then the
    growing degree days are:

    .. math::

        GD4_j = \sum_{i=1}^I (TG_{ij}-{4} | TG_{ij} > {4}℃)
    """
    thresh = utils.convert_units_to(thresh, tas)
    return tas.pipe(lambda x: x - thresh) \
        .clip(min=0) \
        .resample(time=freq) \
        .sum(dim='time')


@declare_units('days', tas='[temperature]', thresh='[temperature]')
def growing_season_length(tas, thresh='5.0 degC', window=6, freq='YS'):
    r"""Growing season length.

    The number of days between the first occurrence of at least
    six consecutive days with mean daily temperature over 5℃ and
    the first occurrence of at least six consecutive days with
    mean daily temperature below 5℃ after July 1st in the northern
    hemisphere and January 1st in the southern hemisphere.

    Parameters
    ---------
    tas : xarray.DataArray
      Mean daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to base evaluation [℃] or [K]. Default: '5.0 degC'.
    window : int
      Minimum number of days with temperature above threshold to mark the beginning and end of growing season.
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Growing season length.

    Notes
    -----
    Let :math:`TG_{ij}` be the mean temperature at day :math:`i` of period :math:`j`. Then counted is
    the number of days between the first occurrence of at least 6 consecutive days with:

    .. math::

        TG_{ij} > 5 ℃

    and the first occurrence after 1 July of at least 6 consecutive days with:

    .. math::

        TG_{ij} < 5 ℃
    """

    # i = xr.DataArray(np.arange(tas.time.size), dims='time')
    # ind = xr.broadcast(i, tas)[0]
    #
    # c = ((tas > thresh) * 1).rolling(time=window).sum()
    # i1 = ind.where(c == window).resample(time=freq).min(dim='time')
    #
    # # Resample sets the time to T00:00.
    # i11 = i1.reindex_like(c, method='ffill')
    #
    # # TODO: Adjust for southern hemisphere
    #
    # #i2 = ind.where(c == 0).where(tas.time.dt.month >= 7)
    # # add check to make sure indice of end of growing season is after growing season start
    # i2 = ind.where((c==0) & (ind > i11)).where(tas.time.dt.month >= 7)
    #
    # d = i2 - i11
    #
    # # take min value (first occurence after july)
    # gsl = d.resample(time=freq).min(dim='time')
    #
    # # turn nan into 0
    # gsl = xr.where(np.isnan(gsl), 0, gsl)

    # compute growth season length on resampled data
    thresh = utils.convert_units_to(thresh, tas)

    c = ((tas > thresh) * 1).rolling(time=window).sum().chunk(tas.chunks)

    def compute_gsl(c):
        nt = c.time.size
        i = xr.DataArray(np.arange(nt), dims='time').chunk({'time': 1})
        ind = xr.broadcast(i, c)[0].chunk(c.chunks)
        i1 = ind.where(c == window).min(dim='time')
        i1 = xr.where(np.isnan(i1), nt, i1)
        i11 = i1.reindex_like(c, method='ffill')
        i2 = ind.where((c == 0) & (ind > i11)).where(c.time.dt.month >= 7)
        i2 = xr.where(np.isnan(i2), nt, i2)
        d = (i2 - i1).min(dim='time')
        return d

    gsl = c.resample(time=freq).apply(compute_gsl)

    return gsl


@declare_units('days', tasmax='[temperature]', thresh='[temperature]')
def heat_wave_index(tasmax, thresh='25.0 degC', window=5, freq='YS'):
    r"""Heat wave index.

    Number of days that are part of a heatwave, defined as five or more consecutive days over 25℃.

    Parameters
    ----------
    tasmax : xarrray.DataArray
      Maximum daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to designate a heatwave [℃] or [K]. Default: '25.0 degC'.
    window : int
      Minimum number of days with temperature above threshold to qualify as a heatwave.
    freq : str, optional
      Resampling frequency

    Returns
    -------
    DataArray
      Heat wave index.
    """
    thresh = utils.convert_units_to(thresh, tasmax)
    over = tasmax > thresh
    group = over.resample(time=freq)

    return group.apply(rl.windowed_run_count, window=window, dim='time')


@declare_units('C days', tas='[temperature]', thresh='[temperature]')
def heating_degree_days(tas, thresh='17.0 degC', freq='YS'):
    r"""Heating degree days

    Sum of degree days below the temperature threshold at which spaces are heated.

    Parameters
    ----------
    tas : xarray.DataArray
      Mean daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to base evaluation [℃] or [K]. Default: '17.0 degC'.
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Heating degree days index.

    Notes
    -----
    Let :math:`TG_{ij}` be the daily mean temperature at day :math:`i` of period :math:`j`. Then the
    heating degree days are:

    .. math::

        HD17_j = \sum_{i=1}^{I} (17℃ - TG_{ij})
    """
    thresh = utils.convert_units_to(thresh, tas)

    return tas.pipe(lambda x: thresh - x) \
        .clip(0) \
        .resample(time=freq) \
        .sum(dim='time')


@declare_units('days', tasmin='[temperature]', thresh='[temperature]')
def tn_days_below(tasmin, thresh='-10.0 degC', freq='YS'):
    r"""Number of days with tmin below a threshold in

    Number of days where daily minimum temperature is below a threshold.

    Parameters
    ----------
    tasmin : xarray.DataArray
      Minimum daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to base evaluation [℃] or [K] . Default: '-10 degC'.
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Number of days Tmin < threshold.

    Notes
    -----
    Let :math:`TN_{ij}` be the daily minimum temperature at day :math:`i` of period :math:`j`. Then
    counted is the number of days where:

    .. math::

        TX_{ij} < Threshold [℃]
    """
    thresh = utils.convert_units_to(thresh, tasmin)
    f1 = utils.threshold_count(tasmin, '<', thresh, freq)
    return f1


@declare_units('days', tasmax='[temperature]', thresh='[temperature]')
def tx_days_above(tasmax, thresh='25.0 degC', freq='YS'):
    r"""Number of summer days

    Number of days where daily maximum temperature exceed a threshold.

    Parameters
    ----------
    tasmax : xarray.DataArray
      Maximum daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to base evaluation [℃] or [K]. Default: '25 degC'.
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Number of summer days.

    Notes
    -----
    Let :math:`TX_{ij}` be the daily maximum temperature at day :math:`i` of period :math:`j`. Then
    counted is the number of days where:

    .. math::

        TX_{ij} > Threshold [℃]
    """
    thresh = utils.convert_units_to(thresh, tasmax)
    f = (tasmax > (thresh)) * 1
    return f.resample(time=freq).sum(dim='time')


@declare_units('days', tasmax='[temperature]', thresh='[temperature]')
def warm_day_frequency(tasmax, thresh='30 degC', freq='YS'):
    r"""Frequency of extreme warm days

    Return the number of days with tasmax > thresh per period

    Parameters
    ----------
    tasmax : xarray.DataArray
      Mean daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to base evaluation [℃] or [K]. Default : '30 degC'
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Number of days exceeding threshold.

    Notes:
    Let :math:`TX_{ij}` be the daily maximum temperature at day :math:`i` of period :math:`j`. Then
    counted is the number of days where:

    .. math::

        TN_{ij} > Threshold [℃]

    """
    thresh = utils.convert_units_to(thresh, tasmax)
    events = (tasmax > thresh) * 1
    return events.resample(time=freq).sum(dim='time')


@declare_units('days', tasmin='[temperature]', thresh='[temperature]')
def warm_night_frequency(tasmin, thresh='22 degC', freq='YS'):
    r"""Frequency of extreme warm nights

    Return the number of days with tasmin > thresh per period

    Parameters
    ----------
    tasmin : xarray.DataArray
      Minimum daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to base evaluation [℃] or [K]. Default : '22 degC'
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      The number of days with tasmin > thresh per period
    """
    thresh = utils.convert_units_to(thresh, tasmin, )
    events = (tasmin > thresh) * 1
    return events.resample(time=freq).sum(dim='time')


@declare_units('days', pr='[precipitation]', thresh='[precipitation]')
def wetdays(pr, thresh='1.0 mm/day', freq='YS'):
    r"""Wet days

    Return the total number of days during period with precipitation over threshold.

    Parameters
    ----------
    pr : xarray.DataArray
      Daily precipitation [mm]
    thresh : str
      Precipitation value over which a day is considered wet. Default: '1 mm/day'.
    freq : str, optional
      Resampling frequency defining the periods
      defined in http://pandas.pydata.org/pandas-docs/stable/timeseries.html#resampling.

    Returns
    -------
    xarray.DataArray
      The number of wet days for each period [day]

    Examples
    --------
    The following would compute for each grid cell of file `pr.day.nc` the number days
    with precipitation over 5 mm at the seasonal frequency, ie DJF, MAM, JJA, SON, DJF, etc.:

    >>> pr = xr.open_dataset('pr.day.nc')
    >>> wd = wetdays(pr, pr_min = 5., freq="QS-DEC")
    """
    thresh = utils.convert_units_to(thresh, pr, 'hydro')

    wd = (pr >= thresh) * 1
    return wd.resample(time=freq).sum(dim='time')


@declare_units('days', pr='[precipitation]', thresh='[precipitation]')
def maximum_consecutive_dry_days(pr, thresh='1 mm/day', freq='YS'):
    r"""Maximum number of consecutive dry days

    Return the maximum number of consecutive days within the period where precipitation
    is below a certain threshold.

    Parameters
    ----------
    pr : xarray.DataArray
      Mean daily precipitation flux [mm]
    thresh : str
      Threshold precipitation on which to base evaluation [mm]. Default : '1 mm/day'
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      The maximum number of consecutive dry days.

    Notes
    -----
    Let :math:`\mathbf{p}=p_0, p_1, \ldots, p_n` be a daily precipitation series and :math:`thresh` the threshold
    under which a day is considered dry. Then let :math:`\mathbf{s}` be the sorted vector of indices :math:`i` where
    :math:`[p_i < thresh] \neq [p_{i+1} < thresh]`, that is, the days when the temperature crosses the threshold.
    Then the maximum number of consecutive dry days is given by

    .. math::

       \max(\mathbf{d}) \quad \mathrm{where} \quad d_j = (s_j - s_{j-1}) [p_{s_j} > thresh]

    where :math:`[P]` is 1 if :math:`P` is true, and 0 if false. Note that this formula does not handle sequences at
    the start and end of the series, but the numerical algorithm does.
    """
    t = utils.convert_units_to(thresh, pr, 'hydro')
    group = (pr < t).resample(time=freq)

    return group.apply(rl.longest_run, dim='time')


@declare_units('mm', pr='[precipitation]')
def max_n_day_precipitation_amount(pr, window=1, freq='YS'):
    r"""Highest precipitation amount cumulated over a n-day moving window.

    Calculate the n-day rolling sum of the original daily total precipitation series
    and determine the maximum value over each period.

    Parameters
    ----------
    da : xarray.DataArray
      Daily precipitation values [Kg m-2 s-1] or [mm]
    window : int
      Window size in days.
    freq : str, optional
      Resampling frequency : default 'YS' (yearly)

    Returns
    -------
    xarray.DataArray
      The highest cumulated n-day precipitation value at the given time frequency.

    Examples
    --------
    The following would compute for each grid cell of file `pr.day.nc` the highest 5-day total precipitation
    at an annual frequency:

    >>> da = xr.open_dataset('pr.day.nc').pr
    >>> window = 5
    >>> output = max_n_day_precipitation_amount(da, window, freq="YS")
    """

    # rolling sum of the values
    arr = pr.rolling(time=window, center=False).sum()
    out = arr.resample(time=freq).max(dim='time', keep_attrs=True)

    out.attrs['units'] = pr.units
    # Adjust values and units to make sure they are daily
    return utils.pint_multiply(out, 1 * units.day, 'mm')


@declare_units('days', tasmin='[temperature]', thresh='[temperature]')
def tropical_nights(tasmin, thresh='20.0 degC', freq='YS'):
    r"""Tropical nights

    The number of days with minimum daily temperature above threshold.

    Parameters
    ----------
    tasmin : xarray.DataArray
      Minimum daily temperature [℃] or [K]
    thresh : str
      Threshold temperature on which to base evaluation [℃] or [K]. Default: '20 degC'.
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Number of days with minimum daily temperature above threshold.

    Notes
    -----
    Let :math:`TN_{ij}` be the daily minimum temperature at day :math:`i` of period :math:`j`. Then
    counted is the number of days where:

    .. math::

        TN_{ij} > Threshold [℃]
    """
    thresh = utils.convert_units_to(thresh, tasmin)
    return tasmin.pipe(lambda x: (tasmin > thresh) * 1) \
        .resample(time=freq) \
        .sum(dim='time')



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from xclim import checks
from xclim.atmos import tg_mean
from xclim.testing.common import tas_series, tasmin_series

TAS_SERIES = tas_series
TASMIN_SERIES = tasmin_series

K2C = 273.15

TESTS_HOME = Path(__file__).absolute().parent
TESTS_DATA = Path(TESTS_HOME, 'testdata')


class TestDateHandling:

    def test_assert_daily(self):
        n = 365  # one day short of a full year
        times = pd.date_range('2000-01-01', freq='1D', periods=n)
        da = xr.DataArray(np.arange(n), [('time', times)], attrs={'units': 'K'})
        tg_mean(da)

    # Bad frequency
    def test_bad_frequency(self):
        with pytest.raises(ValueError):
            n = 365
            times = pd.date_range('2000-01-01', freq='12H', periods=n)
            da = xr.DataArray(np.arange(n), [('time', times)], attrs={'units': 'K'})
            tg_mean(da)

    # Missing one day between the two years
    def test_missing_one_day_between_two_years(self):
        with pytest.raises(ValueError):
            n = 365
            times = pd.date_range('2000-01-01', freq='1D', periods=n)
            times = times.append(pd.date_range('2001-01-01', freq='1D', periods=n))
            da = xr.DataArray(np.arange(2 * n), [('time', times)], attrs={'units': 'K'})
            tg_mean(da)

    # Duplicate dates
    def test_duplicate_dates(self):
        with pytest.raises(ValueError):
            n = 365
            times = pd.date_range('2000-01-01', freq='1D', periods=n)
            times = times.append(pd.date_range('2000-12-29', freq='1D', periods=n))
            da = xr.DataArray(np.arange(2 * n), [('time', times)], attrs={'units': 'K'})
            tg_mean(da)


class TestMissingAnyFills:

    def test_missing_days(self, tas_series):
        a = np.arange(360.)
        a[5:10] = np.nan
        ts = tas_series(a)
        out = checks.missing_any(ts, freq='MS')
        assert out[0]
        assert not out[1]

    def test_missing_months(self):
        n = 66
        times = pd.date_range('2001-12-30', freq='1D', periods=n)
        da = xr.DataArray(np.arange(n), [('time', times)])
        miss = checks.missing_any(da, 'MS')
        np.testing.assert_array_equal(miss, [True, False, False, True])

    def test_missing_years(self):
        n = 378
        times = pd.date_range('2001-12-31', freq='1D', periods=n)
        da = xr.DataArray(np.arange(n), [('time', times)])
        miss = checks.missing_any(da, 'YS')
        np.testing.assert_array_equal(miss, [True, False, True])

    def test_missing_season(self):
        n = 378
        times = pd.date_range('2001-12-31', freq='1D', periods=n)
        da = xr.DataArray(np.arange(n), [('time', times)])
        miss = checks.missing_any(da, 'Q-NOV')
        np.testing.assert_array_equal(miss, [True, False, False, False, True])

    def test_to_period_start(self, tasmin_series):
        a = np.zeros(365) + K2C + 5.0
        a[2] -= 20
        ts = tasmin_series(a)
        miss = checks.missing_any(ts, freq='AS-JUL')
        np.testing.assert_equal(miss, [False])

    def test_to_period_end(self, tasmin_series):
        a = np.zeros(365) + K2C + 5.0
        a[2] -= 20
        ts = tasmin_series(a)
        miss = checks.missing_any(ts, freq='A-JUN')
        np.testing.assert_equal(miss, [False])

    def test_month(self, tasmin_series):
        ts = tasmin_series(np.zeros(36))
        miss = checks.missing_any(ts, freq='YS', month=7)
        np.testing.assert_equal(miss, [False])

        miss = checks.missing_any(ts, freq='YS', month=8)
        np.testing.assert_equal(miss, [True])

        with pytest.raises(ValueError, match=r"No data for selected period."):
            miss = checks.missing_any(ts, freq='YS', month=1)

        miss = checks.missing_any(ts, freq='YS', month=[7, 8])
        np.testing.assert_equal(miss, [True])

        ts = tasmin_series(np.zeros(76))
        miss = checks.missing_any(ts, freq='YS', month=[7, 8])
        np.testing.assert_equal(miss, [False])

    def test_season(self, tasmin_series):
        ts = tasmin_series(np.zeros(360))
        miss = checks.missing_any(ts, freq='YS', season='MAM')
        np.testing.assert_equal(miss, [False])

        miss = checks.missing_any(ts, freq='YS', season='JJA')
        np.testing.assert_array_equal(miss, [True, True])

        miss = checks.missing_any(ts, freq='YS', season='SON')
        np.testing.assert_equal(miss, [False])

    def test_hydro(self):
        fn = Path(TESTS_DATA, 'Raven', 'q_sim.nc')
        ds = xr.open_dataset(fn)
        miss = checks.missing_any(ds.q_sim, freq='YS')
        np.testing.assert_array_equal(miss[:-1], False)
        np.testing.assert_array_equal(miss[-1], True)

        miss = checks.missing_any(ds.q_sim, freq='YS', season='JJA')
        np.testing.assert_array_equal(miss, False)



import datetime as dt
from functools import wraps
from warnings import warn
import logging
import numpy as np
import pandas as pd
import xarray as xr

logging.captureWarnings(True)


# Dev notes
# ---------
#
# I suggest we use `check` for weak checking, and `assert` for strong checking.
# Weak checking would log problems in a log, while strong checking would raise an error.
#
# `functools.wraps` is used to copy the docstring and the function's original name from the source
# function to the decorated function. This allows sphinx to correctly find and document functions.


# TODO: Implement pandas infer_freq in xarray with CFTimeIndex.

def check_valid(var, key, expected):
    r"""Check that a variable's attribute has the expected value. Warn user otherwise."""

    att = getattr(var, key, None)
    if att is None:
        e = 'Variable does not have a `{}` attribute.'.format(key)
        warn(e)
    elif att != expected:
        e = 'Variable has a non-conforming {}. Got `{}`, expected `{}`'.format(key, att, expected)
        warn(e)


def assert_daily(var):
    r"""Assert that the series is daily and monotonic (no jumps in time index).

    A ValueError is raised otherwise."""

    t0, t1 = var.time[:2]

    # This won't work for non-standard calendars. Needs to be implemented in xarray. Comment for now
    if isinstance(t0.values, np.datetime64):
        if pd.infer_freq(var.time.to_pandas()) != 'D':
            raise ValueError("time series is not recognized as daily.")

    # Check that the first time step is one day.
    if np.timedelta64(dt.timedelta(days=1)) != (t1 - t0).data:
        raise ValueError("time series is not daily.")

    # Check that the series has the same time step throughout
    if not var.time.to_pandas().is_monotonic_increasing:
        raise ValueError("time index is not monotonically increasing.")


def check_valid_temperature(var, units):
    r"""Check that variable is air temperature."""

    check_valid(var, 'standard_name', 'air_temperature')
    check_valid(var, 'units', units)
    assert_daily(var)


def check_valid_discharge(var):
    r"""Check that the variable is a discharge."""
    #
    check_valid(var, 'standard_name', 'water_volume_transport_in_river_channel')
    check_valid(var, 'units', 'm3 s-1')


def valid_daily_min_temperature(comp, units='K'):
    r"""Decorator to check that a computation runs on a valid temperature dataset."""

    @wraps(comp)
    def func(tasmin, *args, **kwds):
        check_valid_temperature(tasmin, units)
        check_valid(tasmin, 'cell_methods', 'time: minimum within days')
        return comp(tasmin, **kwds)

    return func


def valid_daily_mean_temperature(comp, units='K'):
    r"""Decorator to check that a computation runs on a valid temperature dataset."""

    @wraps(comp)
    def func(tas, *args, **kwds):
        check_valid_temperature(tas, units)
        check_valid(tas, 'cell_methods', 'time: mean within days')
        return comp(tas, *args, **kwds)

    return func


def valid_daily_max_temperature(comp, units='K'):
    r"""Decorator to check that a computation runs on a valid temperature dataset."""

    @wraps(comp)
    def func(tasmax, *args, **kwds):
        check_valid_temperature(tasmax, units)
        check_valid(tasmax, 'cell_methods', 'time: maximum within days')
        return comp(tasmax, *args, **kwds)

    return func


def valid_daily_max_min_temperature(comp, units='K'):
    r"""Decorator to check that a computation runs on valid min and max temperature datasets."""

    @wraps(comp)
    def func(tasmax, tasmin, **kwds):
        valid_daily_max_temperature(tasmax, units)
        valid_daily_min_temperature(tasmin, units)

        return comp(tasmax, tasmin, **kwds)

    return func


def valid_daily_mean_discharge(comp):
    r"""Decorator to check that a computation runs on valid discharge data."""

    @wraps(comp)
    def func(q, **kwds):
        check_valid_discharge(q)
        return comp(q, **kwds)

    return func


def valid_missing_data_threshold(comp, threshold=0):
    r"""Check that the relative number of missing data points does not exceed a threshold."""
    # TODO
    pass


def check_is_dataarray(comp):
    r"""Decorator to check that a computation has an instance of xarray.DataArray
     as first argument."""

    @wraps(comp)
    def func(data_array, *args, **kwds):
        assert isinstance(data_array, xr.DataArray)
        return comp(data_array, *args, **kwds)

    return func


# This function can probably be made simpler once CFPeriodIndex is implemented.
def missing_any(da, freq, **indexer):
    r"""Return a boolean DataArray indicating whether there are missing days in the resampled array.

    Parameters
    ----------
    da : DataArray
      Input array at daily frequency.
    freq : str
      Resampling frequency.
    **indexer : {dim: indexer, }, optional
      Time attribute and values over which to subset the array. For example, use season='DJF' to select winter values,
      month=1 to select January, or month=[6,7,8] to select summer months. If not indexer is given, all values are
      considered.

    Returns
    -------
    out : DataArray
      A boolean array set to True if any month or year has missing values.
    """
    from . import generic

    if '-' in freq:
        pfreq, anchor = freq.split('-')
    else:
        pfreq = freq

    # Compute the number of days in the time series during each period at the given frequency.
    selected = generic.select_time(da, **indexer)
    if selected.time.size == 0:
        raise ValueError("No data for selected period.")

    c = selected.notnull().resample(time=freq).sum(dim='time')

    # Otherwise simply use the start and end dates to find the expected number of days.
    if pfreq.endswith('S'):
        start_time = c.indexes['time']
        end_time = start_time.shift(1, freq=freq)
    else:
        end_time = c.indexes['time']
        start_time = end_time.shift(-1, freq=freq)

    if indexer:
        # Create a full synthetic time series and compare the number of days with the original series.
        t0 = str(start_time[0].date())
        t1 = str(end_time[-1].date())
        if isinstance(c.indexes['time'], xr.CFTimeIndex):
            cal = da.time.encoding.get('calendar')
            t = xr.cftime_range(t0, t1, freq='D', calendar=cal)
        else:
            t = pd.date_range(t0, t1, freq='D')

        sda = xr.DataArray(data=np.empty(len(t)), coords={'time': t}, dims=('time', ))
        st = generic.select_time(sda, **indexer)
        sn = st.notnull().resample(time=freq).sum(dim='time')
        miss = sn != c
        return miss

    n = (end_time - start_time).days
    nda = xr.DataArray(n.values, coords={'time': c.time}, dims='time')
    return c != nda



import pytest
from xclim import build_module


class TestBuildModules():

    def test_nonexistent_process_build_failure(self):
        name = ""
        objs = {'k1': 'v1', 'k2': 'v2'}
        with pytest.raises(AttributeError):
            build_module(name, objs, mode='warn')

    def test_quiet_build_failure(self):
        name = None
        objs = {}
        with pytest.raises(TypeError):
            build_module(name, objs, mode='ignore')

    def test_loud_build_failure(self):
        name = ""
        objs = {'k1': None, 'k2': None}
        with pytest.warns(Warning):
            build_module(name, objs, mode='warn')

    def test_raise_build_failure(self):
        name = ""
        objs = {'k1': None, 'k2': None}
        with pytest.raises(NotImplementedError):
            build_module(name, objs, mode='raise')


class TestICCLIM():

    def test_exists(self):
        from xclim import icclim
        assert getattr(icclim, 'TG', None) is not None



import pandas as pd
import pytest
import xarray as xr


@pytest.fixture
def tas_series():
    def _tas_series(values, start='7/1/2000'):
        coords = pd.date_range(start, periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time', name='tas',
                            attrs={'standard_name': 'air_temperature',
                                   'cell_methods': 'time: mean within days',
                                   'units': 'K'})

    return _tas_series


@pytest.fixture
def tasmax_series():
    def _tasmax_series(values, start='7/1/2000'):
        coords = pd.date_range(start, periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time', name='tasmax',
                            attrs={'standard_name': 'air_temperature',
                                   'cell_methods': 'time: maximum within days',
                                   'units': 'K'})

    return _tasmax_series


@pytest.fixture
def tasmin_series():
    def _tasmin_series(values, start='7/1/2000'):
        coords = pd.date_range(start, periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time', name='tasmin',
                            attrs={'standard_name': 'air_temperature',
                                   'cell_methods': 'time: minimum within days',
                                   'units': 'K'})

    return _tasmin_series


@pytest.fixture
def pr_series():
    def _pr_series(values, start='7/1/2000'):
        coords = pd.date_range(start, periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time', name='pr',
                            attrs={'standard_name': 'precipitation_flux',
                                   'cell_methods': 'time: sum over day',
                                   'units': 'kg m-2 s-1'})

    return _pr_series



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# Tests for `xclim` package.
#
# We want to tests multiple things here:
#  - that data results are correct
#  - that metadata is correct and complete
#  - that missing data are handled appropriately
#  - that various calendar formats and supported
#  - that non-valid input frequencies or holes in the time series are detected
#
#
# For correctness, I think it would be useful to use a small dataset and run the original ICCLIM indicators on it,
# saving the results in a reference netcdf dataset. We could then compare the hailstorm output to this reference as
# a first line of defense.


# import cftime
import calendar
import os

import numpy as np
import pandas as pd
import pytest
import xarray as xr

import xclim.indices as xci
from xclim.testing.common import tas_series, tasmax_series, tasmin_series, pr_series
from xclim.utils import percentile_doy

xr.set_options(enable_cftimeindex=True)

TAS_SERIES = tas_series
TASMAX_SERIES = tasmax_series
TASMIN_SERIES = tasmin_series
PR_SERIES = pr_series
TESTS_HOME = os.path.abspath(os.path.dirname(__file__))
TESTS_DATA = os.path.join(TESTS_HOME, 'testdata')
K2C = 273.15


# PLEASE MAINTAIN ALPHABETICAL ORDER

class TestBaseFlowIndex:

    def test_simple(self, q_series):
        a = np.zeros(365) + 10
        a[10:17] = 1
        q = q_series(a)
        out = xci.base_flow_index(q)
        np.testing.assert_array_equal(out, 1. / a.mean())


class TestMaxNDayPrecipitationAmount:

    # test 2 day max precip
    def test_single_max(self, pr_series):
        a = pr_series(np.array([3, 4, 20, 20, 0, 6, 9, 25, 0, 0]))
        rxnday = xci.max_n_day_precipitation_amount(a, 2)
        assert rxnday == 40 * 3600 * 24
        assert rxnday.time.dt.year == 2000

    # test whether sum over entire length is resolved
    def test_sumlength_max(self, pr_series):
        a = pr_series(np.array([3, 4, 20, 20, 0, 6, 9, 25, 0, 0]))
        rxnday = xci.max_n_day_precipitation_amount(a, len(a))
        assert rxnday == a.sum('time') * 3600 * 24
        assert rxnday.time.dt.year == 2000

    # test whether non-unique maxes are resolved
    def test_multi_max(self, pr_series):
        a = pr_series(np.array([3, 4, 20, 20, 0, 6, 15, 25, 0, 0]))
        rxnday = xci.max_n_day_precipitation_amount(a, 2)
        assert rxnday == 40 * 3600 * 24
        assert len(rxnday) == 1
        assert rxnday.time.dt.year == 2000


class TestMax1DayPrecipitationAmount:

    @staticmethod
    def time_series(values):
        coords = pd.date_range('7/1/2000', periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time',
                            attrs={'standard_name': 'precipitation_flux',
                                   'cell_methods': 'time: sum (interval: 1 day)',
                                   'units': 'mm/day'})

    # test max precip
    def test_single_max(self):
        a = self.time_series(np.array([3, 4, 20, 0, 0]))
        rx1day = xci.max_1day_precipitation_amount(a)
        assert rx1day == 20
        assert rx1day.time.dt.year == 2000

    # test whether repeated maxes are resolved
    def test_multi_max(self):
        a = self.time_series(np.array([20, 4, 20, 20, 0]))
        rx1day = xci.max_1day_precipitation_amount(a)
        assert rx1day == 20
        assert rx1day.time.dt.year == 2000
        assert len(rx1day) == 1

    # test whether uniform maxes are resolved
    def test_uniform_max(self):
        a = self.time_series(np.array([20, 20, 20, 20, 20]))
        rx1day = xci.max_1day_precipitation_amount(a)
        assert rx1day == 20
        assert rx1day.time.dt.year == 2000
        assert len(rx1day) == 1


class TestColdSpellDurationIndex:

    def test_simple(self, tasmin_series):
        i = 3650
        A = 10.
        tn = np.zeros(i) + A * np.sin(np.arange(i) / 365. * 2 * np.pi) + .1 * np.random.rand(i)
        tn[10:20] -= 2
        tn = tasmin_series(tn)
        tn10 = percentile_doy(tn, per=.1)

        out = xci.cold_spell_duration_index(tn, tn10, freq='YS')
        assert out[0] == 10
        assert out.units == 'days'


class TestColdSpellDays:

    def test_simple(self, tas_series):
        a = np.zeros(365)
        a[10:20] -= 15  # 10 days
        a[40:43] -= 50  # too short -> 0
        a[80:100] -= 30  # at the end and beginning
        da = tas_series(a + K2C)

        out = xci.cold_spell_days(da, thresh='-10. C', freq='M')
        np.testing.assert_array_equal(out, [10, 0, 12, 8, 0, 0, 0, 0, 0, 0, 0, 0])
        assert out.units == 'days'


class TestConsecutiveFrostDays:

    @staticmethod
    def time_series(values):
        coords = pd.date_range('7/1/2000', periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time',
                            attrs={'standard_name': 'air_temperature',
                                   'cell_methods': 'time: minimum within days',
                                   'units': 'K'})

    def test_one_freeze_day(self):
        a = self.time_series(np.array([3, 4, 5, -1, 3]) + K2C)
        cfd = xci.consecutive_frost_days(a)
        assert cfd == 1
        assert cfd.time.dt.year == 2000

    def test_no_freeze(self):
        a = self.time_series(np.array([3, 4, 5, 1, 3]) + K2C)
        cfd = xci.consecutive_frost_days(a)
        assert cfd == 0

    def test_all_year_freeze(self):
        a = self.time_series(np.zeros(365) - 10 + K2C)
        cfd = xci.consecutive_frost_days(a)
        assert cfd == 365


class TestCoolingDegreeDays:

    @staticmethod
    def time_series(values):
        coords = pd.date_range('7/1/2000', periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time',
                            attrs={'standard_name': 'air_temperature',
                                   'cell_methods': 'time: mean within days',
                                   'units': 'K'})

    def test_no_cdd(self):
        a = self.time_series(np.array([10, 15, -5, 18]) + K2C)
        cdd = xci.cooling_degree_days(a)
        assert cdd == 0
        assert cdd.units == 'C days'

    def test_cdd(self):
        a = self.time_series(np.array([20, 25, -15, 19]) + K2C)
        cdd = xci.cooling_degree_days(a)
        assert cdd == 10


class TestDailyFreezeThawCycles:

    def test_simple(self, tasmin_series, tasmax_series):
        mn = np.zeros(365)
        mx = np.zeros(365)

        # 5 days in 1st month
        mn[10:20] -= 1
        mx[10:15] += 1

        # 1 day in 2nd month
        mn[40:44] += [1, 1, -1, -1]
        mx[40:44] += [1, -1, 1, -1]

        mn = tasmin_series(mn + K2C)
        mx = tasmax_series(mx + K2C)
        out = xci.daily_freezethaw_cycles(mx, mn, 'M')
        np.testing.assert_array_equal(out[:2], [5, 1])
        np.testing.assert_array_equal(out[2:], 0)


class TestDailyPrIntensity:

    def test_simple(self, pr_series):
        pr = pr_series(np.zeros(365))
        pr[3:8] += [.5, 1, 2, 3, 4]
        out = xci.daily_pr_intensity(pr, thresh='1 kg/m**2/s')
        np.testing.assert_array_equal(out[0], 2.5 * 3600 * 24)

    def test_mm(self, pr_series):
        pr = pr_series(np.zeros(365))
        pr[3:8] += [.5, 1, 2, 3, 4]
        pr.attrs['units'] = 'mm/d'
        out = xci.daily_pr_intensity(pr, thresh='1 mm/day')
        np.testing.assert_array_almost_equal(out[0], 2.5)


class TestFreshetStart:

    def test_simple(self, tas_series):
        tg = np.zeros(365) - 1
        w = 5

        i = 10
        tg[i:i + w - 1] += 6  # too short

        i = 20
        tg[i:i + w] += 6  # ok

        i = 30
        tg[i:i + w + 1] += 6  # Second valid condition, should be ignored.

        tg = tas_series(tg + K2C, start='1/1/2000')
        out = xci.freshet_start(tg, window=w)
        assert out[0] == tg.indexes['time'][20].dayofyear

    def test_no_start(self, tas_series):
        tg = np.zeros(365) - 1
        tg = tas_series(tg, start='1/1/2000')
        out = xci.freshet_start(tg)
        np.testing.assert_equal(out, [np.nan, ])


class TestGrowingDegreeDays:

    def test_simple(self, tas_series):
        a = np.zeros(365)
        a[0] = 5  # default thresh at 4
        da = tas_series(a + K2C)
        assert xci.growing_degree_days(da)[0] == 1


class TestGrowingSeasonLength:
    def test_simple(self, tas_series):
        # test for different growing length

        # generate 5 years of data
        a = np.zeros(366 * 2 + 365 * 3)
        tas = tas_series(a, start='2000/1/1')

        # 2000 : no growing season

        # 2001 : growing season all year
        d1 = '27-12-2000'
        d2 = '31-12-2001'
        buffer = tas.sel(time=slice(d1, d2))
        tas = tas.where(~tas.time.isin(buffer.time), 280)

        # 2002 : growing season in June only
        d1 = '6-1-2002'
        d2 = '6-10-2002'
        buffer = tas.sel(time=slice(d1, d2))
        tas = tas.where(~tas.time.isin(buffer.time), 280)
        #
        # comment:
        # correct answer should be 10 (i.e. there are 10 days
        # with tas > 5 degC) but current definition imposes end
        # of growing season to be equal or later than July 1st.

        # growing season in Aug only
        d1 = '8-1-2003'
        d2 = '8-10-2003'
        buffer = tas.sel(time=slice(d1, d2))
        tas = tas.where(~tas.time.isin(buffer.time), 280)

        # growing season from June to end of July
        d1 = '6-1-2004'
        d2 = '7-31-2004'
        buffer = tas.sel(time=slice(d1, d2))
        tas = tas.where(~tas.time.isin(buffer.time), 280)

        gsl = xci.growing_season_length(tas)
        target = [0, 365, 25, 10, 61]

        np.testing.assert_array_equal(gsl, target)


class TestHeatingDegreeDays:

    def test_simple(self, tas_series):
        a = np.zeros(365) + 17
        a[:7] += [-3, -2, -1, 0, 1, 2, 3]
        da = tas_series(a + K2C)
        out = xci.heating_degree_days(da)
        np.testing.assert_array_equal(out[:1], 6)
        np.testing.assert_array_equal(out[1:], 0)


class TestHeatWaveIndex:

    def test_simple(self, tasmax_series):
        a = np.zeros(365)
        a[10:20] += 30  # 10 days
        a[40:43] += 50  # too short -> 0
        a[80:100] += 30  # at the end and beginning
        da = tasmax_series(a + K2C)

        out = xci.heat_wave_index(da, thresh='25 C', freq='M')
        np.testing.assert_array_equal(out, [10, 0, 12, 8, 0, 0, 0, 0, 0, 0, 0, 0])


class TestHeatWaveFrequency:

    def test_1d(self, tasmax_series, tasmin_series):
        tn = tasmin_series(np.asarray([20, 23, 23, 23, 23, 22, 23, 23, 23, 23]) + K2C)
        tx = tasmax_series(np.asarray([29, 31, 31, 31, 29, 31, 31, 31, 31, 31]) + K2C)

        # some hw
        hwf = xci.heat_wave_frequency(tn, tx, thresh_tasmin='22 C',
                                      thresh_tasmax='30 C')
        np.testing.assert_allclose(hwf.values, 2)
        hwf = xci.heat_wave_frequency(tn, tx, thresh_tasmin='22 C',
                                      thresh_tasmax='30 C', window=4)
        np.testing.assert_allclose(hwf.values, 1)
        # one long hw
        hwf = xci.heat_wave_frequency(tn, tx, thresh_tasmin='10 C',
                                      thresh_tasmax='10 C')
        np.testing.assert_allclose(hwf.values, 1)
        # no hw
        hwf = xci.heat_wave_frequency(tn, tx, thresh_tasmin='40 C',
                                      thresh_tasmax='40 C')
        np.testing.assert_allclose(hwf.values, 0)


class TestHeatWaveMaxLength:

    def test_1d(self, tasmax_series, tasmin_series):
        tn = tasmin_series(np.asarray([20, 23, 23, 23, 23, 22, 23, 23, 23, 23]) + K2C)
        tx = tasmax_series(np.asarray([29, 31, 31, 31, 29, 31, 31, 31, 31, 31]) + K2C)

        # some hw
        hwml = xci.heat_wave_max_length(tn, tx, thresh_tasmin='22 C',
                                        thresh_tasmax='30 C')
        np.testing.assert_allclose(hwml.values, 4)

        # one long hw
        hwml = xci.heat_wave_max_length(tn, tx, thresh_tasmin='10 C',
                                        thresh_tasmax='10 C')
        np.testing.assert_allclose(hwml.values, 10)

        # no hw
        hwml = xci.heat_wave_max_length(tn, tx, thresh_tasmin='40 C',
                                        thresh_tasmax='40 C')
        np.testing.assert_allclose(hwml.values, 0)

        hwml = xci.heat_wave_max_length(tn, tx, thresh_tasmin='22 C',
                                        thresh_tasmax='30 C', window=5)
        np.testing.assert_allclose(hwml.values, 0)


class TestTnDaysBelow:

    def test_simple(self, tasmin_series):
        a = np.zeros(365)
        a[:6] -= [27, 28, 29, 30, 31, 32]  # 2 above 30
        mx = tasmin_series(a + K2C)

        out = xci.tn_days_below(mx, thresh='-10 C')
        np.testing.assert_array_equal(out[:1], [6])
        np.testing.assert_array_equal(out[1:], [0])
        out = xci.tn_days_below(mx, thresh='-30 C')
        np.testing.assert_array_equal(out[:1], [2])
        np.testing.assert_array_equal(out[1:], [0])


class TestTxDaysAbove:

    def test_simple(self, tasmax_series):
        a = np.zeros(365)
        a[:6] += [27, 28, 29, 30, 31, 32]  # 2 above 30
        mx = tasmax_series(a + K2C)

        out = xci.tx_days_above(mx, thresh='30 C')
        np.testing.assert_array_equal(out[:1], [2])
        np.testing.assert_array_equal(out[1:], [0])


class TestLiquidPrecipitationRatio:

    def test_simple(self, pr_series, tas_series):
        pr = np.zeros(100)
        pr[10:20] = 1
        pr = pr_series(pr)

        tas = np.zeros(100)
        tas[:14] -= 20
        tas[14:] += 10
        tas = tas_series(tas + K2C)

        out = xci.liquid_precip_ratio(pr, tas=tas, freq='M')
        np.testing.assert_almost_equal(out[:1], [.6, ])


class TestMaximumConsecutiveDryDays:

    def test_simple(self, pr_series):
        a = np.zeros(365) + 10
        a[5:15] = 0
        pr = pr_series(a)
        out = xci.maximum_consecutive_dry_days(pr, freq='M')
        assert out[0] == 10

    def test_run_start_at_0(self, pr_series):
        a = np.zeros(365) + 10
        a[:10] = 0
        pr = pr_series(a)
        out = xci.maximum_consecutive_dry_days(pr, freq='M')
        assert out[0] == 10


class TestPrecipAccumulation:
    # build test data for different calendar
    time_std = pd.date_range('2000-01-01', '2010-12-31', freq='D')
    da_std = xr.DataArray(time_std.year, coords=[time_std], dims='time', attrs={'units': 'mm d-1'})

    # calendar 365_day and 360_day not tested for now since xarray.resample
    # does not support other calendars than standard
    #
    # units = 'days since 2000-01-01 00:00'
    # time_365 = cftime.num2date(np.arange(0, 10 * 365), units, '365_day')
    # time_360 = cftime.num2date(np.arange(0, 10 * 360), units, '360_day')
    # da_365 = xr.DataArray(np.arange(time_365.size), coords=[time_365], dims='time')
    # da_360 = xr.DataArray(np.arange(time_360.size), coords=[time_360], dims='time')

    def test_simple(self, pr_series):
        pr = np.zeros(100)
        pr[5:10] = 1
        pr = pr_series(pr)

        out = xci.precip_accumulation(pr, freq='M')
        np.testing.assert_array_equal(out[0], 5 * 3600 * 24)

    def test_yearly(self):
        da_std = self.da_std
        out_std = xci.precip_accumulation(da_std)
        target = [(365 + calendar.isleap(y)) * y for y in np.unique(da_std.time.dt.year)]
        np.testing.assert_allclose(out_std.values, target)


class TestRainOnFrozenGround:

    def test_simple(self, tas_series, pr_series):
        tas = np.zeros(30) - 1
        pr = np.zeros(30)

        tas[10] += 5
        pr[10] += 2

        tas = tas_series(tas + K2C)
        pr = pr_series(pr / 3600 / 24)

        out = xci.rain_on_frozen_ground_days(pr, tas, freq='MS')
        assert out[0] == 1

    def test_small_rain(self, tas_series, pr_series):
        tas = np.zeros(30) - 1
        pr = np.zeros(30)

        tas[10] += 5
        pr[10] += .5

        tas = tas_series(tas + K2C)
        pr = pr_series(pr / 3600 / 24)

        out = xci.rain_on_frozen_ground_days(pr, tas, freq='MS')
        assert out[0] == 0

    def test_consecutive_rain(self, tas_series, pr_series):
        tas = np.zeros(30) - 1
        pr = np.zeros(30)

        tas[10:16] += 5
        pr[10:16] += 5

        tas = tas_series(tas + K2C)
        pr = pr_series(pr)

        out = xci.rain_on_frozen_ground_days(pr, tas, freq='MS')
        assert out[0] == 1


class TestTGXN10p:

    def test_tg10p_simple(self, tas_series):
        i = 366
        tas = np.array(range(i))
        tas = tas_series(tas, start='1/1/2000')
        t10 = percentile_doy(tas, per=.1)

        # create cold spell in june
        tas[175:180] = 1

        out = xci.tg10p(tas, t10, freq='MS')
        assert out[0] == 1
        assert out[5] == 5

    def test_tx10p_simple(self, tasmax_series):
        i = 366
        tas = np.array(range(i))
        tas = tasmax_series(tas, start='1/1/2000')
        t10 = percentile_doy(tas, per=.1)

        # create cold spell in june
        tas[175:180] = 1

        out = xci.tx10p(tas, t10, freq='MS')
        assert out[0] == 1
        assert out[5] == 5

    def test_tn10p_simple(self, tas_series):
        i = 366
        tas = np.array(range(i))
        tas = tas_series(tas, start='1/1/2000')
        t10 = percentile_doy(tas, per=.1)

        # create cold spell in june
        tas[175:180] = 1

        out = xci.tn10p(tas, t10, freq='MS')
        assert out[0] == 1
        assert out[5] == 5

    def test_doy_interpolation(self):
        pytest.importorskip('xarray', '0.11.4')

        # Just a smoke test
        fn_clim = os.path.join(TESTS_DATA, 'CanESM2_365day',
                               'tasmin_day_CanESM2_rcp85_r1i1p1_na10kgrid_qm-moving-50bins-detrend_2095.nc')
        fn = os.path.join(TESTS_DATA, 'HadGEM2-CC_360day',
                          'tasmin_day_HadGEM2-CC_rcp85_r1i1p1_na10kgrid_qm-moving-50bins-detrend_2095.nc')

        with xr.open_dataset(fn_clim) as ds:
            t10 = percentile_doy(ds.tasmin.isel(lat=0, lon=0), per=.1)

        with xr.open_dataset(fn) as ds:
            xci.tn10p(ds.tasmin.isel(lat=0, lon=0), t10, freq='MS')


class TestTGXN90p:

    def test_tg90p_simple(self, tas_series):
        i = 366
        tas = np.array(range(i))
        tas = tas_series(tas, start='1/1/2000')
        t90 = percentile_doy(tas, per=.1)

        # create cold spell in june
        tas[175:180] = 1

        out = xci.tg90p(tas, t90, freq='MS')
        assert out[0] == 30
        assert out[1] == 29
        assert out[5] == 25

    def test_tx90p_simple(self, tasmax_series):
        i = 366
        tas = np.array(range(i))
        tas = tasmax_series(tas, start='1/1/2000')
        t90 = percentile_doy(tas, per=.1)

        # create cold spell in june
        tas[175:180] = 1

        out = xci.tx90p(tas, t90, freq='MS')
        assert out[0] == 30
        assert out[1] == 29
        assert out[5] == 25

    def test_tn90p_simple(self, tasmin_series):
        i = 366
        tas = np.array(range(i))
        tas = tasmin_series(tas, start='1/1/2000')
        t90 = percentile_doy(tas, per=.1)

        # create cold spell in june
        tas[175:180] = 1

        out = xci.tn90p(tas, t90, freq='MS')
        assert out[0] == 30
        assert out[1] == 29
        assert out[5] == 25


class TestTxMin:

    @staticmethod
    def time_series(values):
        coords = pd.date_range('7/1/2000', periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time',
                            attrs={'standard_name': 'air_temperature',
                                   'cell_methods': 'time: maximum within days',
                                   'units': 'K'})


class TestTxMean:

    @staticmethod
    def time_series(values):
        coords = pd.date_range('7/1/2000', periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time',
                            attrs={'standard_name': 'air_temperature',
                                   'cell_methods': 'time: maximum within days',
                                   'units': 'K'})

    def test_attrs(self):
        a = self.time_series(np.array([320, 321, 322, 323, 324]))
        txm = xci.tx_mean(a, freq='YS')
        assert txm == 322
        assert txm.units == 'K'

        a = self.time_series(np.array([20, 21, 22, 23, 24]))
        a.attrs['units'] = 'C'
        txm = xci.tx_mean(a, freq='YS')

        assert txm == 22
        assert txm.units == 'C'


class TestTxMax:

    @staticmethod
    def time_series(values):
        coords = pd.date_range('7/1/2000', periods=len(values), freq=pd.DateOffset(days=1))
        return xr.DataArray(values, coords=[coords, ], dims='time',
                            attrs={'standard_name': 'air_temperature',
                                   'cell_methods': 'time: maximum within days',
                                   'units': 'K'})

    def test_simple(self):
        a = self.time_series(np.array([20, 25, -15, 19]))
        txm = xci.tx_max(a, freq='YS')
        assert txm == 25


class TestTgMaxTgMinIndices:

    @staticmethod
    def random_tmax_tmin_setup(length, tasmax_series, tasmin_series):
        max_values = np.random.uniform(-20, 40, length)
        min_values = []
        for i in range(length):
            min_values.append(np.random.uniform(-40, max_values[i]))
        tasmax = tasmax_series(np.add(max_values, K2C))
        tasmin = tasmin_series(np.add(min_values, K2C))
        return tasmax, tasmin

    @staticmethod
    def static_tmax_tmin_setup(tasmax_series, tasmin_series):
        max_values = np.add([22, 10, 35.2, 25.1, 18.9, 12, 16], K2C)
        min_values = np.add([17, 3.5, 22.7, 16, 12.4, 7, 12], K2C)
        tasmax = tasmax_series(max_values)
        tasmin = tasmin_series(min_values)
        return tasmax, tasmin

    # def test_random_daily_temperature_range(self, tasmax_series, tasmin_series):
    #     days = 365
    #     tasmax, tasmin = self.random_tmax_tmin_setup(days, tasmax_series, tasmin_series)
    #     dtr = xci.daily_temperature_range(tasmax, tasmin, freq="YS")
    #
    #     np.testing.assert_array_less(-dtr, [0, 0])
    #     np.testing.assert_allclose([dtr.mean()], [20], atol=10)

    def test_static_daily_temperature_range(self, tasmax_series, tasmin_series):
        tasmax, tasmin = self.static_tmax_tmin_setup(tasmax_series, tasmin_series)
        dtr = xci.daily_temperature_range(tasmax, tasmin, freq="YS")
        output = np.mean(tasmax - tasmin)

        np.testing.assert_equal(dtr, output)

    # def test_random_variable_daily_temperature_range(self, tasmax_series, tasmin_series):
    #     days = 1095
    #     tasmax, tasmin = self.random_tmax_tmin_setup(days, tasmax_series, tasmin_series)
    #     vdtr = xci.daily_temperature_range_variability(tasmax, tasmin, freq="YS")
    #
    #     np.testing.assert_allclose(vdtr.mean(), 20, atol=10)
    #     np.testing.assert_array_less(-vdtr, [0, 0, 0, 0])

    def test_static_variable_daily_temperature_range(self, tasmax_series, tasmin_series):
        tasmax, tasmin = self.static_tmax_tmin_setup(tasmax_series, tasmin_series)
        dtr = xci.daily_temperature_range_variability(tasmax, tasmin, freq="YS")

        np.testing.assert_almost_equal(dtr, 2.667, decimal=3)

    def test_static_extreme_temperature_range(self, tasmax_series, tasmin_series):
        tasmax, tasmin = self.static_tmax_tmin_setup(tasmax_series, tasmin_series)
        etr = xci.extreme_temperature_range(tasmax, tasmin)

        np.testing.assert_array_almost_equal(etr, 31.7)

    def test_uniform_freeze_thaw_cycles(self, tasmax_series, tasmin_series):
        temp_values = np.zeros(365)
        tasmax, tasmin = tasmax_series(temp_values + 5 + K2C), tasmin_series(temp_values - 5 + K2C)
        ft = xci.daily_freezethaw_cycles(tasmax, tasmin, freq="YS")

        np.testing.assert_array_equal([np.sum(ft)], [365])

    def test_static_freeze_thaw_cycles(self, tasmax_series, tasmin_series):
        tasmax, tasmin = self.static_tmax_tmin_setup(tasmax_series, tasmin_series)
        tasmin -= 15
        ft = xci.daily_freezethaw_cycles(tasmax, tasmin, freq="YS")

        np.testing.assert_array_equal([np.sum(ft)], [4])

    # TODO: Write a better random_freezethaw_cycles test
    # def test_random_freeze_thaw_cycles(self):
    #     runs = np.array([])
    #     for i in range(10):
    #         temp_values = np.random.uniform(-30, 30, 365)
    #         tasmax, tasmin = self.tmax_tmin_time_series(temp_values + K2C)
    #         ft = xci.daily_freezethaw_cycles(tasmax, tasmin, freq="YS")
    #         runs = np.append(runs, ft)
    #
    #     np.testing.assert_allclose(np.mean(runs), 120, atol=20)


class TestWarmDayFrequency:

    def test_1d(self, tasmax_series):
        a = np.zeros(35)
        a[25:] = 31
        da = tasmax_series(a + K2C)
        wdf = xci.warm_day_frequency(da, freq='MS')
        np.testing.assert_allclose(wdf.values, [6, 4])
        wdf = xci.warm_day_frequency(da, freq='YS')
        np.testing.assert_allclose(wdf.values, [10])
        wdf = xci.warm_day_frequency(da, thresh='-1 C')
        np.testing.assert_allclose(wdf.values, [35])
        wdf = xci.warm_day_frequency(da, thresh='50 C')
        np.testing.assert_allclose(wdf.values, [0])


class TestWarmNightFrequency:

    def test_1d(self, tasmin_series):
        a = np.zeros(35)
        a[25:] = 23
        da = tasmin_series(a + K2C)
        wnf = xci.warm_night_frequency(da, freq='MS')
        np.testing.assert_allclose(wnf.values, [6, 4])
        wnf = xci.warm_night_frequency(da, freq='YS')
        np.testing.assert_allclose(wnf.values, [10])
        wnf = xci.warm_night_frequency(da, thresh='-1 C')
        np.testing.assert_allclose(wnf.values, [35])
        wnf = xci.warm_night_frequency(da, thresh='50 C')
        np.testing.assert_allclose(wnf.values, [0])


class TestTxTnDaysAbove:

    def test_1d(self, tasmax_series, tasmin_series):
        tn = tasmin_series(np.asarray([20, 23, 23, 23, 23, 22, 23, 23, 23, 23]) + K2C)
        tx = tasmax_series(np.asarray([29, 31, 31, 31, 29, 31, 31, 31, 31, 31]) + K2C)

        wmmtf = xci.tx_tn_days_above(tn, tx)
        np.testing.assert_allclose(wmmtf.values, [7])
        wmmtf = xci.tx_tn_days_above(tn, tx, thresh_tasmax='50 C')
        np.testing.assert_allclose(wmmtf.values, [0])
        wmmtf = xci.tx_tn_days_above(tn, tx, thresh_tasmax='0 C',
                                     thresh_tasmin='0 C')
        np.testing.assert_allclose(wmmtf.values, [10])


class TestWarmSpellDurationIndex:

    def test_simple(self, tasmax_series):
        i = 3650
        A = 10.
        tx = np.zeros(i) + A * np.sin(np.arange(i) / 365. * 2 * np.pi) + .1 * np.random.rand(i)
        tx[10:20] += 2
        tx = tasmax_series(tx)
        tx90 = percentile_doy(tx, per=.9)

        out = xci.warm_spell_duration_index(tx, tx90, freq='YS')
        assert out[0] == 10


class TestWinterRainRatio:

    def test_simple(self, pr_series, tas_series):
        pr = np.ones(450)
        pr = pr_series(pr, start='12/1/2000')

        tas = np.zeros(450) - 1
        tas[10:20] += 10
        tas = tas_series(tas + K2C, start='12/1/2000')

        out = xci.winter_rain_ratio(pr, tas=tas)
        np.testing.assert_almost_equal(out, [10. / (31 + 31 + 28), 0])


# I'd like to parametrize some of these tests so we don't have to write individual tests for each indicator.
class TestTG:
    def test_cmip3(self, cmip3_day_tas):
        pytest.importorskip('xarray', '0.11.4')
        xci.tg_mean(cmip3_day_tas)

    def compare_against_icclim(self, cmip3_day_tas):
        pass


@pytest.fixture(scope="session")
def cmip3_day_tas():
    # xr.set_options(enable_cftimeindex=False)
    ds = xr.open_dataset(os.path.join(TESTS_DATA, 'cmip3', 'tas.sresb1.giss_model_e_r.run1.atm.da.nc'))
    yield ds.tas
    ds.close()


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


@pytest.mark.skip
def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string

# x = Test_frost_days()
# print('done')



import pytest
from xclim import streamflow
import numpy as np


def test_base_flow_index(ndq_series):
    out = streamflow.base_flow_index(ndq_series, freq='YS')
    assert out.attrs['units'] == ''


class Test_FA():

    def test_simple(self, ndq_series):
        out = streamflow.freq_analysis(ndq_series, mode='max', t=[2, 5], dist='gamma', season='DJF')
        assert out.long_name == 'N-year return period max winter 1-day flow'
        assert out.shape == (2, 2, 3)  # nrt, nx, ny

    def test_no_indexer(self, ndq_series):
        out = streamflow.freq_analysis(ndq_series, mode='max', t=[2, 5], dist='gamma')
        assert out.long_name == 'N-year return period max annual 1-day flow'
        assert out.shape == (2, 2, 3)  # nrt, nx, ny

    def test_q27(self, ndq_series):
        out = streamflow.freq_analysis(ndq_series, mode='max', t=2, dist='gamma', window=7)
        assert out.shape == (1, 2, 3)


class TestStats():

    def test_simple(self, ndq_series):
        out = streamflow.stats(ndq_series, freq='YS', op='min', season='MAM')
        assert out.attrs['units'] == 'm^3 s-1'

    def test_missing(self, ndq_series):
        a = ndq_series
        a = ndq_series.where(~((a.time.dt.dayofyear == 5) * (a.time.dt.year == 1902)))
        out = streamflow.stats(a, op='max', month=1)

        np.testing.assert_array_equal(out[1].isnull(), False)
        np.testing.assert_array_equal(out[2].isnull(), True)


class TestFit():

    def test_simple(self, ndq_series):
        ts = streamflow.stats(ndq_series, freq='YS', op='max')
        p = streamflow.fit(ts, dist='gumbel_r')
        assert p.attrs['estimator'] == 'Maximum likelihood'


def test_qdoy_max(ndq_series, q_series):
    out = streamflow.doy_qmax(ndq_series, freq='YS', season='JJA')
    assert out.attrs['units'] == ''

    a = np.ones(450)
    a[100] = 2
    out = streamflow.doy_qmax(q_series(a), freq='YS')
    assert out[0] == 101



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'scipy>=1.2',
    'numpy>=1.15',
    'pandas>=0.23',
    'cftime>=1.0.3',
    'netCDF4>=1.4',
    'dask[complete]',
    'bottleneck>=1.2.1',
    'xarray>=0.12.0',
    'pyproj>=1.9.5.1',
    'pint>=0.8',
    'boltons>=18.0',
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', 'tox', ]

docs_requirements = ['Sphinx', 'guzzle-sphinx-theme', ]

KEYWORDS = "xclim climate climatology netcdf gridded analysis"

setup(
    author="Travis Logan",
    author_email='logan.travis@ouranos.ca',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
    ],
    description="Derived climate variables built with xarray.",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords=KEYWORDS,
    name='xclim',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    extras_require={'docs': docs_requirements},
    url='https://github.com/Ouranosinc/xclim',
    version='0.10.2-beta',
    zip_safe=False,
)



# -*- coding: utf-8 -*-

from xclim import indices
from xclim.utils import Indicator, Indicator2D

__all__ = ['rain_on_frozen_ground_days', 'max_1day_precipitation_amount',
           'max_n_day_precipitation_amount', 'wetdays', 'maximum_consecutive_dry_days', 'maximum_consecutive_wet_days',
           'daily_pr_intensity', 'precip_accumulation']


class Pr(Indicator):
    context = 'hydro'


class PrTas(Indicator2D):
    context = 'hydro'


rain_on_frozen_ground_days = PrTas(identifier='rain_frzgr',
                                   units='days',
                                   standard_name='number_of_days_with_lwe_thickness_of_'
                                                 'precipitation_amount_above_threshold',
                                   long_name='Number of rain on frozen ground days',
                                   description="{freq} number of days with rain above {thresh} "
                                               "after a series of seven days "
                                               "with average daily temperature below 0℃. "
                                               "Precipitation is assumed to be rain when the"
                                               "daily average temperature is above 0℃.",
                                   cell_methods='',
                                   compute=indices.rain_on_frozen_ground_days,
                                   )

max_1day_precipitation_amount = Pr(identifier='rx1day',
                                   units='mm/day',
                                   standard_name='lwe_thickness_of_precipitation_amount',
                                   long_name='maximum 1-day total precipitation',
                                   description="{freq} maximum 1-day total precipitation",
                                   cellmethods='time: sum within days time: maximum over days',
                                   compute=indices.max_1day_precipitation_amount,
                                   )

max_n_day_precipitation_amount = Pr(identifier='max_n_day_precipitation_amount',
                                    var_name='rx{window}day',
                                    units='mm',
                                    standard_name='lwe_thickness_of_precipitation_amount',
                                    long_name='maximum {window}-day total precipitation',
                                    description="{freq} maximum {window}-day total precipitation.",
                                    cellmethods='time: sum within days time: maximum over days',
                                    compute=indices.max_n_day_precipitation_amount,
                                    )

wetdays = Pr(identifier='wetdays',
             units='days',
             standard_name='number_of_days_with_lwe_thickness_of_precipitation_amount_at_or_above_threshold',
             long_name='Number of wet days (precip >= {thresh})',
             description='{freq} number of days with daily precipitation over {thresh}.',
             cell_methods='time: sum within days time: sum over days',
             compute=indices.wetdays,
             )

maximum_consecutive_wet_days = Pr(identifier='cwd',
                                  units='days',
                                  standard_name='number_of_days_with_lwe_thickness_of_'
                                                'precipitation_amount_at_or_above_threshold',
                                  long_name='Maximum consecutive wet days (Precip >= {thresh})',
                                  description='{freq} maximum number of days with daily '
                                              'precipitation over {thresh}.',
                                  cell_methods='time: sum within days time: sum over days',
                                  compute=indices.maximum_consecutive_wet_days,
                                  )

maximum_consecutive_dry_days = Pr(identifier='cdd',
                                  units='days',
                                  standard_name='number_of_days_with_lwe_thickness_of_'
                                                'precipitation_amount_below_threshold',
                                  long_name='Maximum consecutive dry days (Precip < {thresh})',
                                  description='{freq} maximum number of days with daily '
                                              'precipitation below {thresh}.',
                                  cell_methods='time: sum within days time: sum over days',
                                  compute=indices.maximum_consecutive_dry_days,
                                  )

daily_pr_intensity = Pr(identifier='sdii',
                        units='mm/day',
                        standard_name='lwe_thickness_of_precipitation_amount',
                        long_name='Average precipitation during Wet Days (SDII)',
                        description="{freq} Simple Daily Intensity Index (SDII) : {freq} average precipitation "
                                    "for days with daily precipitation over {thresh}.",
                        cell_methods='',
                        compute=indices.daily_pr_intensity,
                        )

precip_accumulation = Pr(identifier='prcptot',
                         units='mm',
                         standard_name='lwe_thickness_of_precipitation_amount',
                         long_name='Total precipitation',
                         description='{freq} total precipitation.',
                         cell_methods='time: sum within days time: sum over days',
                         compute=indices.precip_accumulation
                         )



import logging

import xarray as xr

from xclim import run_length as rl, utils
from xclim.utils import declare_units, units

# logging.basicConfig(level=logging.DEBUG)
# logging.captureWarnings(True)

xr.set_options(enable_cftimeindex=True)  # Set xarray to use cftimeindex


# Frequencies : YS: year start, QS-DEC: seasons starting in december, MS: month start
# See http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases

# -------------------------------------------------- #
# ATTENTION: ASSUME ALL INDICES WRONG UNTIL TESTED ! #
# -------------------------------------------------- #

__all__ = ['tg_max', 'tg_mean', 'tg_min', 'tn_max', 'tn_mean', 'tn_min', 'tx_max', 'tx_mean', 'tx_min',
           'base_flow_index', 'consecutive_frost_days', 'frost_days', 'ice_days', 'max_1day_precipitation_amount',
           'precip_accumulation']


@declare_units('[temperature]', tas='[temperature]')
def tg_max(tas, freq='YS'):
    r"""Highest mean temperature.

    The maximum of daily mean temperature.

    Parameters
    ----------
    tas : xarray.DataArray
      Mean daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Maximum of daily minimum temperature.

    Notes
    -----
    Let :math:`TN_{ij}` be the mean temperature at day :math:`i` of period :math:`j`. Then the maximum
    daily mean temperature for period :math:`j` is:

    .. math::

        TNx_j = max(TN_{ij})
    """

    return tas.resample(time=freq).max(dim='time', keep_attrs=True)


@declare_units('[temperature]', tas='[temperature]')
def tg_mean(tas, freq='YS'):
    r"""Mean of daily average temperature.

    Resample the original daily mean temperature series by taking the mean over each period.

    Parameters
    ----------
    tas : xarray.DataArray
      Mean daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      The mean daily temperature at the given time frequency

    Notes
    -----
    Let :math:`TN_i` be the mean daily temperature of day :math:`i`, then for a period :math:`p` starting at
    day :math:`a` and finishing on day :math:`b`:

    .. math::

       TG_p = \frac{\sum_{i=a}^{b} TN_i}{b - a + 1}


    Examples
    --------
    The following would compute for each grid cell of file `tas.day.nc` the mean temperature
    at the seasonal frequency, ie DJF, MAM, JJA, SON, DJF, etc.:

    >>> t = xr.open_dataset('tas.day.nc')
    >>> tg = tm_mean(t, freq="QS-DEC")
    """

    arr = tas.resample(time=freq) if freq else tas
    return arr.mean(dim='time', keep_attrs=True)


@declare_units('[temperature]', tas='[temperature]')
def tg_min(tas, freq='YS'):
    r"""Lowest mean temperature

    Minimum of daily mean temperature.

    Parameters
    ----------
    tas : xarray.DataArray
      Mean daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Minimum of daily minimum temperature.

    Notes
    -----
    Let :math:`TG_{ij}` be the mean temperature at day :math:`i` of period :math:`j`. Then the minimum
    daily mean temperature for period :math:`j` is:

    .. math::

        TGn_j = min(TG_{ij})
    """

    return tas.resample(time=freq).min(dim='time', keep_attrs=True)


@declare_units('[temperature]', tasmin='[temperature]')
def tn_max(tasmin, freq='YS'):
    r"""Highest minimum temperature.

    The maximum of daily minimum temperature.

    Parameters
    ----------
    tasmin : xarray.DataArray
      Minimum daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Maximum of daily minimum temperature.

    Notes
    -----
    Let :math:`TN_{ij}` be the minimum temperature at day :math:`i` of period :math:`j`. Then the maximum
    daily minimum temperature for period :math:`j` is:

    .. math::

        TNx_j = max(TN_{ij})
    """

    return tasmin.resample(time=freq).max(dim='time', keep_attrs=True)


@declare_units('[temperature]', tasmin='[temperature]')
def tn_mean(tasmin, freq='YS'):
    r"""Mean minimum temperature.

    Mean of daily minimum temperature.

    Parameters
    ----------
    tasmin : xarray.DataArray
      Minimum daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Mean of daily minimum temperature.

    Notes
    -----
    Let :math:`TN_{ij}` be the minimum temperature at day :math:`i` of period :math:`j`. Then mean
    values in period :math:`j` are given by:

    .. math::

        TN_{ij} = \frac{ \sum_{i=1}^{I} TN_{ij} }{I}
    """

    arr = tasmin.resample(time=freq) if freq else tasmin
    return arr.mean(dim='time', keep_attrs=True)


@declare_units('[temperature]', tasmin='[temperature]')
def tn_min(tasmin, freq='YS'):
    r"""Lowest minimum temperature

    Minimum of daily minimum temperature.

    Parameters
    ----------
    tasmin : xarray.DataArray
      Minimum daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Minimum of daily minimum temperature.

    Notes
    -----
    Let :math:`TN_{ij}` be the minimum temperature at day :math:`i` of period :math:`j`. Then the minimum
    daily minimum temperature for period :math:`j` is:

    .. math::

        TNn_j = min(TN_{ij})
    """

    return tasmin.resample(time=freq).min(dim='time', keep_attrs=True)


@declare_units('[temperature]', tasmax='[temperature]')
def tx_max(tasmax, freq='YS'):
    r"""Highest max temperature

    The maximum value of daily maximum temperature.

    Parameters
    ----------
    tasmax : xarray.DataArray
      Maximum daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Maximum value of daily maximum temperature.

    Notes
    -----
    Let :math:`TX_{ij}` be the maximum temperature at day :math:`i` of period :math:`j`. Then the maximum
    daily maximum temperature for period :math:`j` is:

    .. math::

        TXx_j = max(TX_{ij})
    """

    return tasmax.resample(time=freq).max(dim='time', keep_attrs=True)


@declare_units('[temperature]', tasmax='[temperature]')
def tx_mean(tasmax, freq='YS'):
    r"""Mean max temperature

    The mean of daily maximum temperature.

    Parameters
    ----------
    tasmax : xarray.DataArray
      Maximum daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Mean of daily maximum temperature.

    Notes
    -----
    Let :math:`TX_{ij}` be the maximum temperature at day :math:`i` of period :math:`j`. Then mean
    values in period :math:`j` are given by:

    .. math::

        TX_{ij} = \frac{ \sum_{i=1}^{I} TX_{ij} }{I}
    """

    arr = tasmax.resample(time=freq) if freq else tasmax
    return arr.mean(dim='time', keep_attrs=True)


@declare_units('[temperature]', tasmax='[temperature]')
def tx_min(tasmax, freq='YS'):
    r"""Lowest max temperature

    The minimum of daily maximum temperature.

    Parameters
    ----------
    tasmax : xarray.DataArray
      Maximum daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Minimum of daily maximum temperature.

    Notes
    -----
    Let :math:`TX_{ij}` be the maximum temperature at day :math:`i` of period :math:`j`. Then the minimum
    daily maximum temperature for period :math:`j` is:

    .. math::

        TXn_j = min(TX_{ij})
    """

    return tasmax.resample(time=freq).min(dim='time', keep_attrs=True)


@declare_units('', q='[discharge]')
def base_flow_index(q, freq='YS'):
    r"""Base flow index

    Return the base flow index, defined as the minimum 7-day average flow divided by the mean flow.

    Parameters
    ----------
    q : xarray.DataArray
      Rate of river discharge [m³/s]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArrray
      Base flow index.

    Notes
    -----
    Let :math:`\mathbf{q}=q_0, q_1, \ldots, q_n` be the sequence of daily discharge and :math:`\overline{\mathbf{q}}`
    the mean flow over the period. The base flow index is given by:

    .. math::

       \frac{\min(\mathrm{CMA}_7(\mathbf{q}))}{\overline{\mathbf{q}}}


    where :math:`\mathrm{CMA}_7` is the seven days moving average of the daily flow:

    .. math::

       \mathrm{CMA}_7(q_i) = \frac{\sum_{j=i-3}^{i+3} q_j}{7}

    """

    m7 = q.rolling(time=7, center=True).mean().resample(time=freq)
    mq = q.resample(time=freq)

    m7m = m7.min(dim='time')
    return m7m / mq.mean(dim='time')


@declare_units('days', tasmin='[temperature]')
def consecutive_frost_days(tasmin, freq='AS-JUL'):
    r"""Maximum number of consecutive frost days (Tmin < 0℃).

    Resample the daily minimum temperature series by computing the maximum number
    of days below the freezing point over each period.

    Parameters
    ----------
    tasmin : xarray.DataArray
      Minimum daily temperature values [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      The maximum number of consecutive days below the freezing point.

    Notes
    -----
    Let :math:`\mathbf{x}=x_0, x_1, \ldots, x_n` be a daily minimum temperature series and
    :math:`\mathbf{s}` be the sorted vector of indices :math:`i` where :math:`[p_i < 0\celsius] \neq [p_{i+1} <
    0\celsius]`, that is, the days when the temperature crosses the freezing point.
    Then the maximum number of consecutive frost days is given by

    .. math::

       \max(\mathbf{d}) \quad \mathrm{where} \quad d_j = (s_j - s_{j-1}) [x_{s_j} > 0\celsius]

    where :math:`[P]` is 1 if :math:`P` is true, and 0 if false. Note that this formula does not handle sequences at
    the start and end of the series, but the numerical algorithm does.
    """
    tu = units.parse_units(tasmin.attrs['units'].replace('-', '**-'))
    fu = 'degC'
    frz = 0
    if fu != tu:
        frz = units.convert(frz, fu, tu)
    group = (tasmin < frz).resample(time=freq)
    return group.apply(rl.longest_run, dim='time')


@declare_units('days', tasmin='[temperature]')
def frost_days(tasmin, freq='YS'):
    r"""Frost days index

    Number of days where daily minimum temperatures are below 0℃.

    Parameters
    ----------
    tasmin : xarray.DataArray
      Minimum daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Frost days index.

    Notes
    -----
    Let :math:`TN_{ij}` be the daily minimum temperature at day :math:`i` of period :math:`j`. Then
    counted is the number of days where:

    .. math::

        TN_{ij} < 0℃
    """
    tu = units.parse_units(tasmin.attrs['units'].replace('-', '**-'))
    fu = 'degC'
    frz = 0
    if fu != tu:
        frz = units.convert(frz, fu, tu)
    f = (tasmin < frz) * 1
    return f.resample(time=freq).sum(dim='time')


@declare_units('days', tasmax='[temperature]')
def ice_days(tasmax, freq='YS'):
    r"""Number of ice/freezing days

    Number of days where daily maximum temperatures are below 0℃.

    Parameters
    ----------
    tasmax : xarrray.DataArray
      Maximum daily temperature [℃] or [K]
    freq : str, optional
      Resampling frequency

    Returns
    -------
    xarray.DataArray
      Number of ice/freezing days.

    Notes
    -----
    Let :math:`TX_{ij}` be the daily maximum temperature at day :math:`i` of period :math:`j`. Then
    counted is the number of days where:

    .. math::

        TX_{ij} < 0℃
    """
    tu = units.parse_units(tasmax.attrs['units'].replace('-', '**-'))
    fu = 'degC'
    frz = 0
    if fu != tu:
        frz = units.convert(frz, fu, tu)
    f = (tasmax < frz) * 1
    return f.resample(time=freq).sum(dim='time')


@declare_units('mm/day', pr='[precipitation]')
def max_1day_precipitation_amount(pr, freq='YS'):
    r"""Highest 1-day precipitation amount for a period (frequency).

    Resample the original daily total precipitation temperature series by taking the max over each period.

    Parameters
    ----------
    pr : xarray.DataArray
      Daily precipitation values [Kg m-2 s-1] or [mm]
    freq : str, optional
      Resampling frequency one of : 'YS' (yearly) ,'M' (monthly), or 'QS-DEC' (seasonal - quarters starting in december)

    Returns
    -------
    xarray.DataArray
      The highest 1-day precipitation value at the given time frequency.

    Notes
    -----
    Let :math:`PR_i` be the mean daily precipitation of day `i`, then for a period `j`:

    .. math::

       PRx_{ij} = max(PR_{ij})

    Examples
    --------
    The following would compute for each grid cell of file `pr.day.nc` the highest 1-day total
    at an annual frequency:

    >>> pr = xr.open_dataset('pr.day.nc').pr
    >>> rx1day = max_1day_precipitation_amount(pr, freq="YS")
    """

    out = pr.resample(time=freq).max(dim='time', keep_attrs=True)
    return utils.convert_units_to(out, 'mm/day', 'hydro')


@declare_units('mm', pr='[precipitation]')
def precip_accumulation(pr, freq='YS'):
    r"""Accumulated total (liquid + solid) precipitation.

    Resample the original daily mean precipitation flux and accumulate over each period.

    Parameters
    ----------
    pr : xarray.DataArray
      Mean daily precipitation flux [Kg m-2 s-1] or [mm].
    freq : str, optional
      Resampling frequency as defined in
      http://pandas.pydata.org/pandas-docs/stable/timeseries.html#resampling.

    Returns
    -------
    xarray.DataArray
      The total daily precipitation at the given time frequency.

    Notes
    -----
    Let :math:`PR_i` be the mean daily precipitation of day :math:`i`, then for a period :math:`j` starting at
    day :math:`a` and finishing on day :math:`b`:

    .. math::

       PR_{ij} = \sum_{i=a}^{b} PR_i

    Examples
    --------
    The following would compute for each grid cell of file `pr_day.nc` the total
    precipitation at the seasonal frequency, ie DJF, MAM, JJA, SON, DJF, etc.:

    >>> pr_day = xr.open_dataset('pr_day.nc').pr
    >>> prcp_tot_seasonal = precip_accumulation(pr_day, freq="QS-DEC")
    """

    out = pr.resample(time=freq).sum(dim='time', keep_attrs=True)
    return utils.pint_multiply(out, 1 * units.day, 'mm')



# -*- coding: utf-8 -*-
# Note: stats.dist.shapes: comma separated names of shape parameters
# The other parameters, common to all distribution, are loc and scale.

import dask
import numpy as np
import xarray as xr


def select_time(da, **indexer):
    """Select entries according to a time period.

    Parameters
    ----------
    da : xarray.DataArray
      Input data.
    **indexer : {dim: indexer, }, optional
      Time attribute and values over which to subset the array. For example, use season='DJF' to select winter values,
      month=1 to select January, or month=[6,7,8] to select summer months. If not indexer is given, all values are
      considered.

    Returns
    -------
    xr.DataArray
      Selected input values.
    """
    if not indexer:
        selected = da
    else:
        key, val = indexer.popitem()
        time_att = getattr(da.time.dt, key)
        selected = da.sel(time=time_att.isin(val)).dropna(dim='time')

    return selected


def select_resample_op(da, op, freq="YS", **indexer):
    """Apply operation over each period that is part of the index selection.

    Parameters
    ----------
    da : xarray.DataArray
      Input data.
    op : str {'min', 'max', 'mean', 'std', 'var', 'count', 'sum', 'argmax', 'argmin'} or func
      Reduce operation. Can either be a DataArray method or a function that can be applied to a DataArray.
    freq : str
      Resampling frequency defining the periods
      defined in http://pandas.pydata.org/pandas-docs/stable/timeseries.html#resampling.
    **indexer : {dim: indexer, }, optional
      Time attribute and values over which to subset the array. For example, use season='DJF' to select winter values,
      month=1 to select January, or month=[6,7,8] to select summer months. If not indexer is given, all values are
      considered.

    Returns
    -------
    xarray.DataArray
      The maximum value for each period.
    """
    da = select_time(da, **indexer)
    r = da.resample(time=freq, keep_attrs=True)
    if isinstance(op, str):
        return getattr(r, op)(dim='time', keep_attrs=True)

    return r.apply(op)


def doymax(da):
    """Return the day of year of the maximum value."""
    i = da.argmax(dim='time')
    out = da.time.dt.dayofyear[i]
    out.attrs['units'] = ''
    return out


def doymin(da):
    """Return the day of year of the minimum value."""
    i = da.argmax(dim='time')
    out = da.time.dt.dayofyear[i]
    out.attrs['units'] = ''
    return out


def fit(da, dist='norm'):
    """Fit an array to a univariate distribution along the time dimension.

    Parameters
    ----------
    da : xarray.DataArray
      Time series to be fitted along the time dimension.
    dist : str
      Name of the univariate distribution, such as beta, expon, genextreme, gamma, gumbel_r, lognorm, norm
      (see scipy.stats).

    Returns
    -------
    xarray.DataArray
      An array of distribution parameters fitted using the method of Maximum Likelihood.

    Notes
    -----
    Coordinates for which all values are NaNs will be dropped before fitting the distribution. If the array
    still contains NaNs, the distribution parameters will be returned as NaNs.
    """
    # Get the distribution
    dc = get_dist(dist)

    # Fit the parameters (lazy computation)
    data = dask.array.apply_along_axis(dc.fit, da.get_axis_num('time'), da.dropna('time', how='all'))

    # Count the number of values used for the fit.
    # n = arr.count(dim='time')

    # Create a view to a DataArray with the desired dimensions to copy them over to the parameter array.
    mean = da.mean(dim='time', keep_attrs=True)

    # Create coordinate for the distribution parameters
    coords = dict(mean.coords.items())
    coords['dparams'] = ([] if dc.shapes is None else dc.shapes.split(',')) + ['loc', 'scale']

    # TODO: add time and time_bnds coordinates (Low will work on this)
    # time.attrs['climatology'] = 'climatology_bounds'
    # coords['time'] =
    # coords['climatology_bounds'] =

    out = xr.DataArray(data=data, coords=coords, dims=(u'dparams',) + mean.dims)
    out.attrs = da.attrs
    out.attrs['original_name'] = getattr(da, 'standard_name', '')
    out.attrs['description'] = \
        'Parameters of the {0} distribution fitted over {1}'.format(dist, getattr(da, 'standard_name', ''))
    out.attrs['estimator'] = 'Maximum likelihood'
    out.attrs['scipy_dist'] = dist
    out.attrs['units'] = ''
    # out.name = 'params'
    return out


def fa(da, t, dist='norm', mode='high'):
    """Return the value corresponding to the given return period.

    Parameters
    ----------
    da : xarray.DataArray
      Maximized/minimized input data with a `time` dimension.
    t : int or sequence
      Return period. The period depends on the resolution of the input data. If the input array's resolution is
      yearly, then the return period is in years.
    dist : str
      Name of the univariate distribution, such as beta, expon, genextreme, gamma, gumbel_r, lognorm, norm
      (see scipy.stats).
    mode : {'min', 'max}
      Whether we are looking for a probability of exceedance (max) or a probability of non-exceedance (min).

    Returns
    -------
    xarray.DataArray
      An array of values with a 1/t probability of exceedance (if mode=='max').
    """
    t = np.atleast_1d(t)

    # Get the distribution
    dc = get_dist(dist)

    # Fit the parameters of the distribution
    p = fit(da, dist)

    # Create a lambda function to facilitate passing arguments to dask. There is probably a better way to do this.
    if mode in ['max', 'high']:
        def func(x):
            return dc.isf(1. / t, *x)
    elif mode in ['min', 'low']:
        def func(x):
            return dc.ppf(1. / t, *x)
    else:
        raise ValueError("mode `{}` should be either 'max' or 'min'".format(mode))

    data = dask.array.apply_along_axis(func, p.get_axis_num('dparams'), p)

    # Create coordinate for the return periods
    coords = dict(p.coords.items())
    coords.pop('dparams')
    coords['return_period'] = t

    # Create dimensions
    dims = list(p.dims)
    dims.remove('dparams')
    dims.insert(0, u'return_period')

    # TODO: add time and time_bnds coordinates (Low will work on this)
    # time.attrs['climatology'] = 'climatology_bounds'
    # coords['time'] =
    # coords['climatology_bounds'] =

    out = xr.DataArray(data=data, coords=coords, dims=dims)
    out.attrs = p.attrs
    out.attrs['standard_name'] = '{0} quantiles'.format(dist)
    out.attrs['long_name'] = '{0} return period values for {1}'.format(dist, getattr(da, 'standard_name', ''))
    out.attrs['cell_methods'] = (out.attrs.get('cell_methods', '') + ' dparams: ppf').strip()
    out.attrs['units'] = da.attrs.get('units', '')
    out.attrs['mode'] = mode
    out.attrs['history'] = out.attrs.get('history', '') + "Compute values corresponding to return periods."

    return out


def frequency_analysis(da, mode, t, dist, window=1, freq=None, **indexer):
    """Return the value corresponding to a return period.

    Parameters
    ----------
    da : xarray.DataArray
      Input data.
    t : int or sequence
      Return period. The period depends on the resolution of the input data. If the input array's resolution is
      yearly, then the return period is in years.
    dist : str
      Name of the univariate distribution, such as beta, expon, genextreme, gamma, gumbel_r, lognorm, norm
      (see scipy.stats).
    mode : {'min', 'max'}
      Whether we are looking for a probability of exceedance (high) or a probability of non-exceedance (low).
    window : int
      Averaging window length (days).
    freq : str
      Resampling frequency. If None, the frequency is assumed to be 'YS' unless the indexer is season='DJF',
      in which case `freq` would be set to `YS-DEC`.
    **indexer : {dim: indexer, }, optional
      Time attribute and values over which to subset the array. For example, use season='DJF' to select winter values,
      month=1 to select January, or month=[6,7,8] to select summer months. If not indexer is given, all values are
      considered.

    Returns
    -------
    xarray.DataArray
      An array of values with a 1/t probability of exceedance or non-exceedance when mode is high or low respectively.

    """
    # Apply rolling average
    attrs = da.attrs.copy()
    if window > 1:
        da = da.rolling(time=window, center=False).mean()
        da.attrs.update(attrs)

    # Assign default resampling frequency if not provided
    freq = freq or default_freq(**indexer)

    # Extract the time series of min or max over the period
    sel = select_resample_op(da, op=mode, freq=freq, **indexer).dropna(dim='time')

    # Frequency analysis
    return fa(sel, t, dist, mode)


def default_freq(**indexer):
    """Return the default frequency."""
    freq = 'AS-JAN'
    if indexer:
        if 'DJF' in indexer.values():
            freq = 'AS-DEC'
        if 'month' in indexer and sorted(indexer.values()) != indexer.values():
            raise (NotImplementedError)

    return freq


def get_dist(dist):
    """Return a distribution object from scipy.stats.
    """
    from scipy import stats

    dc = getattr(stats, dist, None)
    if dc is None:
        e = "Statistical distribution `{}` is not in scipy.stats.".format(dist)
        raise ValueError(e)
    return dc
